@@ -8,9 +8,10 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 from collections.abc import MutableMapping, MutableSequence
-from typing import TYPE_CHECKING, cast, Any, Dict, Union
+from typing import TYPE_CHECKING, Any, Dict, Union
 
-from .default import ElementData, XMLSchemaConverter
+from ..exceptions import XMLSchemaTypeError, XMLSchemaValueError
+from .default import ElementData, stackable, XMLSchemaConverter
 
 if TYPE_CHECKING:
     from ..validators import XsdElement
@@ -28,6 +29,7 @@ class UnorderedConverter(XMLSchemaConverter):
     """
     __slots__ = ()
 
+    @stackable
     def element_encode(self, obj: Any, xsd_element: 'XsdElement', level: int = 0) -> ElementData:
         """
         Extracts XML decoded data from a data structure for encoding into an ElementTree.
@@ -37,23 +39,22 @@ class UnorderedConverter(XMLSchemaConverter):
         :param level: the level related to the encoding process (0 means the root).
         :return: an ElementData instance.
         """
-        if level:
-            tag = xsd_element.name
+        if level or not self.preserve_root:
+            element_name = None
+        elif not isinstance(obj, MutableMapping):
+            raise XMLSchemaTypeError(f"A dictionary expected, got {type(obj)} instead.")
+        elif len(obj) != 1:
+            raise XMLSchemaValueError("The dictionary must have exactly one element.")
         else:
-            tag = xsd_element.qualified_name
-            if self.preserve_root and isinstance(obj, MutableMapping):
-                match_local_name = cast(bool, self.strip_namespaces or self.default_namespace)
-                match = xsd_element.get_matching_item(obj, self.ns_prefix, match_local_name)
-                if match is not None:
-                    obj = match
+            element_name, obj = next(iter(obj.items()))
 
         if not isinstance(obj, MutableMapping):
             if xsd_element.type.simple_type is not None:
-                return ElementData(tag, obj, None, {})
+                return ElementData(xsd_element.name, obj, None, {}, None)
             elif xsd_element.type.mixed and isinstance(obj, (str, bytes)):
-                return ElementData(tag, None, [(1, obj)], {})
+                return ElementData(xsd_element.name, None, [(1, obj)], {}, None)
             else:
-                return ElementData(tag, None, obj, {})
+                return ElementData(xsd_element.name, None, obj, {}, None)
 
         text = None
         attributes = {}
@@ -63,57 +64,51 @@ class UnorderedConverter(XMLSchemaConverter):
         # building content_lu, content which is not a list or lists to be placed
         # into a single element (element has a list content type) must be wrapped
         # in a list to retain that structure. Character data are not wrapped into
-        # lists because they because they are divided from the rest of the content
-        # into the unordered mode generator function of the ModelVisitor class.
+        # lists because they are divided from the rest of the content into the
+        # unordered mode generator function of the ModelVisitor class.
         content_lu: Dict[Union[int, str], Any] = {}
 
+        xmlns = self.set_context(obj, level)
+
+        if element_name is None:
+            tag = xsd_element.name
+        else:
+            tag = self.unmap_qname(element_name)
+            if not xsd_element.is_matching(tag, self.default_namespace):
+                raise XMLSchemaValueError("data tag does not match XSD element name")
+
         for name, value in obj.items():
             if name == self.text_key:
                 text = value
             elif self.cdata_prefix is not None and \
                     name.startswith(self.cdata_prefix) and \
-                    name[len(self.cdata_prefix):].isdigit():
-                index = int(name[len(self.cdata_prefix):])
-                content_lu[index] = value
-            elif name == self.ns_prefix:
-                self[''] = value
-            elif name.startswith(f'{self.ns_prefix}:'):
-                self[name[len(self.ns_prefix) + 1:]] = value
+                    (index := name[len(self.cdata_prefix):]).isdigit():
+                content_lu[int(index)] = value
+            elif self.is_xmlns(name):
+                continue
             elif self.attr_prefix and \
                     name.startswith(self.attr_prefix) and \
-                    name != self.attr_prefix:
-                attr_name = name[len(self.attr_prefix):]
+                    (attr_name := name[len(self.attr_prefix):]):
                 ns_name = self.unmap_qname(attr_name, xsd_element.attributes)
                 attributes[ns_name] = value
             elif not isinstance(value, MutableSequence) or not value:
-                content_lu[self.unmap_qname(name)] = [value]
+                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value))
+                content_lu[ns_name] = [value]
             elif isinstance(value[0], (MutableMapping, MutableSequence)):
-                content_lu[self.unmap_qname(name)] = value
+                ns_name = self.unmap_qname(name, xmlns=self.get_xmlns_from_data(value[0]))
+                content_lu[ns_name] = value
             else:
-                xsd_group = xsd_element.type.model_group
-                if xsd_group is None:
-                    xsd_group = xsd_element.any_type.model_group
-                    assert xsd_group is not None
-
                 # `value` is a list but not a list of lists or list of dicts.
                 ns_name = self.unmap_qname(name)
-                for xsd_child in xsd_group.iter_elements():
-                    matched_element = xsd_child.match(ns_name, resolve=True)
-                    if matched_element is not None:
-                        if matched_element.type and matched_element.type.is_list():
-                            content_lu[self.unmap_qname(name)] = [value]
-                        else:
-                            content_lu[self.unmap_qname(name)] = value
-                        break
-                else:
-                    if self.attr_prefix == '' and ns_name not in attributes:
-                        for xsd_attribute in xsd_element.attributes.values():
-                            if xsd_attribute.is_matching(ns_name):
-                                attributes[ns_name] = value
-                                break
-                        else:
-                            content_lu[self.unmap_qname(name)] = [value]
+                xsd_child = xsd_element.match_child(ns_name)
+                if xsd_child is not None:
+                    if xsd_child.type and xsd_child.type.is_list():
+                        content_lu[ns_name] = [value]
                     else:
-                        content_lu[self.unmap_qname(name)] = [value]
+                        content_lu[ns_name] = value
+                elif self.attr_prefix == '' and ns_name in xsd_element.attributes:
+                    attributes[ns_name] = value
+                else:
+                    content_lu[ns_name] = value
 
-        return ElementData(tag, text, content_lu, attributes)
+        return ElementData(tag, text, content_lu, attributes, xmlns)
