@@ -121,8 +121,8 @@ class XsdAttribute(XsdComponent, ValidationMixin[str, DecodedValueType]):
 
                 if self.parent is None or self.qualified:
                     if self.target_namespace == XSI_NAMESPACE and \
-                            name not in {'nil', 'type', 'schemaLocation',
-                                         'noNamespaceSchemaLocation'}:
+                            name not in ('nil', 'type', 'schemaLocation',
+                                         'noNamespaceSchemaLocation'):
                         msg = _("cannot add attributes in %r namespace")
                         self.parse_error(msg % XSI_NAMESPACE)
                     self.name = get_qname(self.target_namespace, name)
@@ -317,7 +317,7 @@ class Xsd11Attribute(XsdAttribute):
             msg = _("attribute 'fixed' with use=prohibited is not allowed in XSD 1.1")
             self.parse_error(msg)
         if 'inheritable' in self.elem.attrib:
-            if self.elem.attrib['inheritable'].strip() in {'true', '1'}:
+            if self.elem.attrib['inheritable'].strip() in ('true', '1'):
                 self.inheritable = True
         self._parse_target_namespace()
 
@@ -656,7 +656,8 @@ class XsdAttributeGroup(
                 yield from attr.iter_components(xsd_classes)
 
     def iter_decode(self, obj: MutableMapping[str, str], validation: str = 'lax',
-                    **kwargs: Any) -> IterDecodeType[List[Tuple[str, Any]]]:
+                    use_defaults: bool = True, **kwargs: Any) \
+            -> IterDecodeType[List[Tuple[str, Any]]]:
         if not obj and not self:
             return
 
@@ -664,11 +665,10 @@ class XsdAttributeGroup(
             reason = _("missing required attribute {!r}").format(name)
             yield self.validation_error(validation, reason, obj, **kwargs)
 
-        kwargs['level'] = kwargs.get('level', 0) + 1
         try:
-            use_defaults = kwargs['use_defaults']
+            kwargs['level'] += 1
         except KeyError:
-            use_defaults = True
+            kwargs['level'] = 1
 
         additional_attrs = [
             (k, v) for k, v in self.iter_value_constraints(use_defaults) if k not in obj
@@ -734,7 +734,8 @@ class XsdAttributeGroup(
         yield result_list
 
     def iter_encode(self, obj: MutableMapping[str, Any], validation: str = 'lax',
-                    **kwargs: Any) -> IterEncodeType[List[Tuple[str, Union[str, List[str]]]]]:
+                    use_defaults: bool = True, **kwargs: Any) \
+            -> IterEncodeType[List[Tuple[str, Union[str, List[str]]]]]:
         if not obj and not self:
             return
 
@@ -742,11 +743,6 @@ class XsdAttributeGroup(
             reason = _("missing required attribute {!r}").format(name)
             yield self.validation_error(validation, reason, obj, **kwargs)
 
-        try:
-            use_defaults = kwargs['use_defaults']
-        except KeyError:
-            use_defaults = True
-
         result_list = []
         for name, value in obj.items():
             try:
