@@ -27,7 +27,7 @@ from .helpers import get_xsd_derivation_attribute
 from .xsdbase import XSD_TYPE_DERIVATIONS, XsdComponent, XsdType, ValidationMixin
 from .attributes import XsdAttributeGroup
 from .assertions import XsdAssert
-from .simple_types import FacetsValueType, XsdSimpleType, XsdUnion
+from .simple_types import FacetsValueType, XsdSimpleType, XsdUnion, XsdAtomic
 from .groups import XsdGroup
 from .wildcards import XsdOpenContent, XsdDefaultOpenContent
 
@@ -88,7 +88,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
                 self._block = kwargs['block']
             if 'final' in kwargs:
                 self._final = kwargs['final']
-        super(XsdComplexType, self).__init__(elem, schema, parent, name)
+        super().__init__(elem, schema, parent, name)
         assert self.content is not None
 
     def __repr__(self) -> str:
@@ -107,7 +107,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
             return  # a local restriction is already parsed by the caller
 
         if 'abstract' in self.elem.attrib:
-            if self.elem.attrib['abstract'].strip() in {'true', '1'}:
+            if self.elem.attrib['abstract'].strip() in ('true', '1'):
                 self.abstract = True
 
         if 'block' in self.elem.attrib:
@@ -123,7 +123,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
                 self.parse_error(err)
 
         if 'mixed' in self.elem.attrib:
-            if self.elem.attrib['mixed'].strip() in {'true', '1'}:
+            if self.elem.attrib['mixed'].strip() in ('true', '1'):
                 self.mixed = True
 
         try:
@@ -536,19 +536,22 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
         else:
             return 'element-only'
 
+    @property
+    def root_type(self) -> BaseXsdType:
+        if self.attributes or self.base_type is None:
+            return cast('XsdComplexType', self.maps.types[XSD_ANY_TYPE])
+        else:
+            return self.base_type.root_type
+
     @property
     def sequence_type(self) -> str:
         if self.is_empty():
             return 'empty-sequence()'
-        elif not self.has_simple_content():
-            st = 'xs:untypedAtomic'
+        elif isinstance(self.content, XsdAtomic):
+            name = self.content.primitive_type.local_name
+            st = 'item()' if name is None else f'xs:{name}'
         else:
-            try:
-                name = self.content.primitive_type.local_name  # type: ignore[union-attr]
-            except AttributeError:
-                st = 'xs:untypedAtomic'
-            else:
-                st = 'item()' if name is None else f'xs:{name}'
+            st = 'xs:untypedAtomic'
 
         return f"{st}{'*' if self.is_emptiable() else '+'}"
 
@@ -602,6 +605,10 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
     def is_list(self) -> bool:
         return isinstance(self.content, XsdSimpleType) and self.content.is_list()
 
+    def is_dynamic_consistent(self, other: Any) -> bool:
+        return other.name == XSD_ANY_TYPE or self.is_derived(other) or \
+            isinstance(other, XsdUnion) and any(self.is_derived(mt) for mt in other.member_types)
+
     def validate(self, obj: Union[ElementType, str, bytes],
                  use_defaults: bool = True,
                  namespaces: Optional[NamespacesType] = None,
@@ -614,7 +621,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
             'extra_validator': extra_validator
         }
         if not isinstance(obj, (str, bytes)):
-            super(XsdComplexType, self).validate(obj, **kwargs)
+            super().validate(obj, **kwargs)
         elif isinstance(self.content, XsdSimpleType):
             self.content.validate(obj, **kwargs)
         elif not self.mixed and self.base_type is not None:
@@ -632,7 +639,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
             'extra_validator': extra_validator
         }
         if not isinstance(obj, (str, bytes)):
-            return super(XsdComplexType, self).is_valid(obj, **kwargs)
+            return super().is_valid(obj, **kwargs)
         elif isinstance(self.content, XsdSimpleType):
             return self.content.is_valid(obj, **kwargs)
         else:
@@ -709,7 +716,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
     def decode(self, obj: Union[ElementType, str, bytes], *args: Any, **kwargs: Any) \
             -> DecodeType[Any]:
         if not isinstance(obj, (str, bytes)):
-            return super(XsdComplexType, self).decode(obj, *args, **kwargs)
+            return super().decode(obj, *args, **kwargs)
         elif isinstance(self.content, XsdSimpleType):
             return self.content.decode(obj, *args, **kwargs)
         else:
@@ -763,9 +770,7 @@ class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], An
             value = obj
 
         xsd_type: BaseXsdType
-        if isinstance(value, list):
-            xsd_type = self.any_simple_type
-        elif isinstance(value, AnyAtomicType):
+        if isinstance(value, AnyAtomicType):
             xsd_type = self.any_atomic_type
         else:
             xsd_type = self.any_type
@@ -829,7 +834,7 @@ class Xsd11ComplexType(XsdComplexType):
             return self.schema.default_open_content
 
     def _parse(self) -> None:
-        super(Xsd11ComplexType, self)._parse()
+        super()._parse()
 
         if self.base_type and self.base_type.base_type is self.any_simple_type and \
                 self.base_type.derivation == 'extension' and not self.attributes:
@@ -841,14 +846,11 @@ class Xsd11ComplexType(XsdComplexType):
         # Add open content to a complex content type
         if isinstance(self.content, XsdGroup):
             if self.open_content is None:
-                if self.content.interleave is not None or self.content.suffix is not None:
+                if self.content.open_content is not None:
                     msg = _("openContent mismatch between type and model group")
                     self.parse_error(msg)
-            elif self.open_content.mode == 'interleave':
-                self.content.interleave = self.content.suffix \
-                    = self.open_content.any_element
-            elif self.open_content.mode == 'suffix':
-                self.content.suffix = self.open_content.any_element
+            elif self.open_content:
+                self.content.open_content = self.open_content
 
         # Add inheritable attributes
         if isinstance(self.base_type, XsdComplexType):
@@ -862,7 +864,7 @@ class Xsd11ComplexType(XsdComplexType):
 
         if 'defaultAttributesApply' not in self.elem.attrib:
             self.default_attributes_apply = True
-        elif self.elem.attrib['defaultAttributesApply'].strip() in {'false', '0'}:
+        elif self.elem.attrib['defaultAttributesApply'].strip() in ('false', '0'):
             self.default_attributes_apply = False
         else:
             self.default_attributes_apply = True
