@@ -16,7 +16,7 @@ from decimal import Decimal
 from types import GeneratorType
 from typing import TYPE_CHECKING, cast, Any, Dict, Iterator, List, Optional, \
     Set, Tuple, Type, Union
-from xml.etree.ElementTree import Element
+from xml.etree.ElementTree import Element, ParseError
 
 from elementpath import XPath2Parser, ElementPathError, XPathContext, XPathToken, \
     ElementNode, LazyElementNode, SchemaElementNode, build_schema_node_tree
@@ -30,22 +30,23 @@ from ..aliases import ElementType, SchemaType, BaseXsdType, SchemaElementType, \
     ModelParticleType, ComponentClassType, AtomicValueType, DecodeType, \
     IterDecodeType, IterEncodeType
 from ..translation import gettext as _
-from ..helpers import get_qname, get_namespace, etree_iter_location_hints, \
-    raw_xml_encode, strictly_equal
+from ..helpers import get_qname, etree_iter_location_hints, \
+    etree_iter_namespaces, raw_xml_encode, strictly_equal
+from ..namespaces import NamespaceMapper
+from ..locations import normalize_url
 from .. import dataobjects
 from ..converters import ElementData, XMLSchemaConverter
-from ..xpath import XsdSchemaProtocol, XsdElementProtocol, XMLSchemaProxy, \
-    ElementPathMixin, XPathElement
+from ..xpath import XMLSchemaProxy, ElementPathMixin, XPathElement
 from ..resources import XMLResource
 
-from .exceptions import XMLSchemaNotBuiltError, XMLSchemaValidationError, \
-    XMLSchemaTypeTableWarning
+from .exceptions import XMLSchemaValidationError, XMLSchemaParseError, \
+    XMLSchemaStopValidation, XMLSchemaTypeTableWarning
 from .helpers import get_xsd_derivation_attribute
 from .xsdbase import XSD_TYPE_DERIVATIONS, XSD_ELEMENT_DERIVATIONS, \
-    XsdComponent, ValidationMixin
+    XSD_VALIDATION_MODES, XsdComponent, ValidationMixin
 from .particles import ParticleMixin, OccursCalculator
 from .identities import XsdIdentity, XsdKey, XsdUnique, \
-    XsdKeyref, KeyrefCounter, IdentityCounterType
+    XsdKeyref, KeyrefCounter, FieldValueSelector
 from .simple_types import XsdSimpleType
 from .attributes import XsdAttribute
 from .wildcards import XsdAnyElement
@@ -105,8 +106,8 @@ class XsdElement(XsdComponent, ParticleMixin,
 
     identities: List[XsdIdentity]
     selected_by: Set[XsdIdentity]
-    alternatives = ()  # type: Union[Tuple[()], List[XsdAlternative]]
-    inheritable = ()  # type: Union[Tuple[()], Dict[str, XsdAttribute]]
+    alternatives: Union[Tuple[()], List['XsdAlternative']] = ()
+    inheritable: Union[Tuple[()], Dict[str, XsdAttribute]] = ()
 
     _ADMITTED_TAGS = {XSD_ELEMENT}
     _block: Optional[str] = None
@@ -116,7 +117,7 @@ class XsdElement(XsdComponent, ParticleMixin,
 
     binding: Optional[DataBindingType] = None
 
-    def __init__(self, elem: Element,
+    def __init__(self, elem: ElementType,
                  schema: SchemaType,
                  parent: Optional[XsdComponent] = None,
                  build: bool = True) -> None:
@@ -124,7 +125,7 @@ class XsdElement(XsdComponent, ParticleMixin,
         if not build:
             self._build = False
         self.selected_by = set()
-        super(XsdElement, self).__init__(elem, schema, parent)
+        super().__init__(elem, schema, parent)
 
     def __repr__(self) -> str:
         return '%s(%s=%r, occurs=%r)' % (
@@ -140,7 +141,7 @@ class XsdElement(XsdComponent, ParticleMixin,
                 self.attributes = self.schema.create_empty_attribute_group(self)
             else:
                 self.attributes = value.attributes
-        super(XsdElement, self).__setattr__(name, value)
+        super().__setattr__(name, value)
 
     def __iter__(self) -> Iterator[SchemaElementType]:
         if self.type.has_complex_content():
@@ -182,8 +183,8 @@ class XsdElement(XsdComponent, ParticleMixin,
                 self.alternatives = xsd_element.alternatives
                 self.selected_by = xsd_element.selected_by
 
-            for attr_name in {'type', 'nillable', 'default', 'fixed', 'form',
-                              'block', 'abstract', 'final', 'substitutionGroup'}:
+            for attr_name in ('type', 'nillable', 'default', 'fixed', 'form',
+                              'block', 'abstract', 'final', 'substitutionGroup'):
                 if attr_name in attrib:
                     msg = _("attribute {!r} is not allowed when element reference is used")
                     self.parse_error(msg.format(attr_name))
@@ -208,7 +209,7 @@ class XsdElement(XsdComponent, ParticleMixin,
             if self.parent is not None:
                 msg = _("local scope elements cannot have abstract attribute")
                 self.parse_error(msg)
-            if attrib['abstract'].strip() in {'true', '1'}:
+            if attrib['abstract'].strip() in ('true', '1'):
                 self.abstract = True
 
         if 'block' in attrib:
@@ -219,7 +220,7 @@ class XsdElement(XsdComponent, ParticleMixin,
             except ValueError as err:
                 self.parse_error(err)
 
-        if 'nillable' in attrib and attrib['nillable'].strip() in {'true', '1'}:
+        if 'nillable' in attrib and attrib['nillable'].strip() in ('true', '1'):
             self.nillable = True
 
         if self.parent is None:
@@ -231,12 +232,12 @@ class XsdElement(XsdComponent, ParticleMixin,
                 except ValueError as err:
                     self.parse_error(err)
 
-            for attr_name in {'ref', 'form', 'minOccurs', 'maxOccurs'}:
+            for attr_name in ('ref', 'form', 'minOccurs', 'maxOccurs'):
                 if attr_name in attrib:
                     msg = _("attribute {!r} is not allowed in a global element declaration")
                     self.parse_error(msg.format(attr_name))
         else:
-            for attr_name in {'final', 'substitutionGroup'}:
+            for attr_name in ('final', 'substitutionGroup'):
                 if attr_name in attrib:
                     msg = _("attribute {!r} not allowed in a local element declaration")
                     self.parse_error(msg.format(attr_name))
@@ -392,20 +393,17 @@ class XsdElement(XsdComponent, ParticleMixin,
 
     @property
     def xpath_proxy(self) -> XMLSchemaProxy:
-        return XMLSchemaProxy(
-            schema=cast(XsdSchemaProtocol, self.schema),
-            base_element=cast(XsdElementProtocol, self)
-        )
+        return XMLSchemaProxy(self.schema, self)
 
     @property
     def xpath_node(self) -> SchemaElementNode:
         schema_node = self.schema.xpath_node
-        node = schema_node.get_element_node(cast(XsdElementProtocol, self))
+        node = schema_node.get_element_node(self)
         if isinstance(node, SchemaElementNode):
             return node
 
         return build_schema_node_tree(
-            root=cast(XsdElementProtocol, self),
+            root=self,
             elements=schema_node.elements,
             global_elements=schema_node.children,
         )
@@ -548,43 +546,39 @@ class XsdElement(XsdComponent, ParticleMixin,
                 return None
         return self.type.text_decode(text)
 
-    def check_dynamic_context(self, elem: ElementType, **kwargs: Any) -> None:
+    def check_dynamic_context(self, elem: ElementType,
+                              validation: str,
+                              options: Dict[str, Any]) -> Iterator[XMLSchemaValidationError]:
         try:
-            locations = kwargs['locations']
+            source: XMLResource = options['source']
         except KeyError:
             return
 
-        schema: Optional[SchemaType]
         for ns, url in etree_iter_location_hints(elem):
-            if ns not in locations:
-                locations[ns] = url
-            elif locations[ns] is None:
+            base_url = source.base_url
+            url = normalize_url(url, base_url)
+            if any(url == schema.url for schema in self.maps.iter_schemas()):
+                continue
+
+            if ns in etree_iter_namespaces(source.root, elem):
                 reason = _("schemaLocation declaration after namespace start")
-                raise XMLSchemaValidationError(self, elem, reason)
+                yield self.validation_error(validation, reason, elem, **options)
 
-            if ns == self.target_namespace:
-                schema = self.schema.include_schema(url, self.schema.base_url)
-            else:
-                schema = self.schema.import_schema(ns, url, self.schema.base_url)
+            try:
+                if ns in self.maps.namespaces:
+                    schema = self.maps.namespaces[ns][0]
+                    schema.include_schema(url)
+                    self.schema.clear()
+                    self.schema.build()
+                else:
+                    self.schema.import_schema(ns, url, base_url, build=True)
 
-            if schema is None:
-                reason = _("missing dynamic loaded schema from %s") % url
-                raise XMLSchemaValidationError(self, elem, reason)
-            elif not schema.built:
-                reason = _("dynamic loaded schema change the assessment")
-                raise XMLSchemaValidationError(self, elem, reason)
-
-        if elem.attrib:
-            for name in elem.attrib:
-                if name[0] == '{':
-                    ns = get_namespace(name)
-                    if ns not in locations:
-                        locations[ns] = None
-
-        if elem.tag[0] == '{':
-            ns = get_namespace(elem.tag)
-            if ns not in locations:
-                locations[ns] = None
+            except (XMLSchemaValidationError, ParseError) as err:
+                yield self.validation_error(validation, err, elem, **options)
+            except XMLSchemaParseError as err:
+                yield self.validation_error(validation, err.message, elem, **options)
+            except OSError:
+                continue
 
     def iter_decode(self, obj: ElementType, validation: str = 'lax', **kwargs: Any) \
             -> IterDecodeType[Any]:
@@ -598,11 +592,35 @@ class XsdElement(XsdComponent, ParticleMixin,
         validation or decoding errors.
         """
         error: Union[XMLSchemaValueError, XMLSchemaValidationError]
+        result: Any
 
         if self.abstract:
-            reason = _("cannot use an abstract element for validation")
-            yield self.validation_error(validation, reason, obj, **kwargs)
+            if self.name == obj.tag:
+                reason = _("can't use an abstract element in an instance")
+                yield self.validation_error(validation, reason, obj, **kwargs)
+            elif self.name not in self.maps.substitution_groups:
+                reason = _("can't use an abstract XSD element for validation "
+                           "unless it's the head of a substitution group")
+                yield self.validation_error(validation, reason, obj, **kwargs)
+            else:
+                for xsd_element in self.iter_substitutes():
+                    if obj.tag == xsd_element.name:
+                        yield from xsd_element.iter_decode(obj, validation, **kwargs)
+                        return
+                else:
+                    reason = _("can't use an abstract XSD element for validation")
+                    yield self.validation_error(validation, reason, obj, **kwargs)
 
+        # Control validation on element and its descendants or stop validation
+        if 'validation_hook' in kwargs:
+            value = kwargs['validation_hook'](obj, self)
+            if value:
+                if isinstance(value, str) and value in XSD_VALIDATION_MODES:
+                    validation = value
+                else:
+                    return
+
+        kwargs['elem'] = obj
         try:
             level = kwargs['level']
         except KeyError:
@@ -622,15 +640,17 @@ class XsdElement(XsdComponent, ParticleMixin,
         try:
             converter = kwargs['converter']
         except KeyError:
-            converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+            converter = self._get_converter(obj, kwargs)
         else:
-            if converter is not None and not isinstance(converter, XMLSchemaConverter):
-                converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+            if not isinstance(converter, NamespaceMapper):
+                converter = self._get_converter(obj, kwargs)
 
-        try:
-            pass  # self.check_dynamic_context(elem, **kwargs) TODO: dynamic schema load
-        except XMLSchemaValidationError as err:
-            yield self.validation_error(validation, err, obj, **kwargs)
+        if not level:
+            # Need to set base context with the right object (the resource can be lazy)
+            converter.set_context(obj, level)
+        elif kwargs.get('use_location_hints'):
+            # Use location hints for dynamic schema load
+            yield from self.check_dynamic_context(obj, validation, options=kwargs)
 
         inherited = kwargs.get('inherited')
         value = content = attributes = None
@@ -641,11 +661,7 @@ class XsdElement(XsdComponent, ParticleMixin,
         if XSI_TYPE in obj.attrib and self.schema.meta_schema is not None:
             # Meta-schema elements ignore xsi:type (issue #350)
             type_name = obj.attrib[XSI_TYPE].strip()
-            try:
-                namespaces = kwargs['namespaces']
-            except KeyError:
-                namespaces = None
-
+            namespaces = converter.namespaces
             try:
                 xsd_type = self.maps.get_instance_type(type_name, xsd_type, namespaces)
             except (KeyError, TypeError) as err:
@@ -654,28 +670,12 @@ class XsdElement(XsdComponent, ParticleMixin,
                 if self.identities:
                     xpath_element = XPathElement(self.name, xsd_type)
                     for identity in self.identities:
-                        if isinstance(identity.elements, tuple) \
-                                or identity.selector is None:
+                        if not identity.built or identity.selector is None:
                             continue  # Skip unbuilt or incomplete identities
-                        elif identity.selector.token is None:
-                            raise XMLSchemaNotBuiltError(
-                                identity, "identity selector is not built"
-                            )
-
-                        context = XPathContext(
-                            root=self.schema.xpath_node,
-                            item=cast(XsdElementProtocol, xpath_element)
+                        identity.elements.update(
+                            identity.get_selected_elements(xpath_element)
                         )
 
-                        for e in identity.selector.token.select_results(context):
-                            if not isinstance(e, (XsdElement, XsdAnyElement)):
-                                reason = _("selector xpath expression can only select elements")
-                                yield self.validation_error(validation, reason, e, **kwargs)
-                            elif e not in identity.elements:
-                                identity.elements[e] = None
-                                if isinstance(e, XsdElement):
-                                    e.selected_by.add(identity)
-
             if xsd_type.is_blocked(self):
                 reason = _("usage of %r is blocked") % xsd_type
                 yield self.validation_error(validation, reason, obj, **kwargs)
@@ -690,7 +690,6 @@ class XsdElement(XsdComponent, ParticleMixin,
 
         # Decode attributes
         attribute_group = self.get_attributes(xsd_type)
-        result: Any
         for result in attribute_group.iter_decode(obj.attrib, validation, **kwargs):
             if isinstance(result, XMLSchemaValidationError):
                 yield self.validation_error(validation, result, obj, **kwargs)
@@ -767,7 +766,7 @@ class XsdElement(XsdComponent, ParticleMixin,
                     reason = _("must have the fixed value %r") % self.fixed
                     yield self.validation_error(validation, reason, obj, **kwargs)
 
-            elif not text and self.default is not None and kwargs.get('use_defaults'):
+            elif not text and self.default is not None and kwargs.get('use_defaults', True):
                 text = self.default
 
             if not isinstance(xsd_type, XsdSimpleType):
@@ -816,8 +815,10 @@ class XsdElement(XsdComponent, ParticleMixin,
                 if not kwargs.get('binary_types'):
                     value = str(value)
 
-        if converter is not None:
-            element_data = ElementData(obj.tag, value, content, attributes)
+        xmlns = converter.set_context(obj, level)  # Purge existing sub-contexts
+
+        if isinstance(converter, XMLSchemaConverter):
+            element_data = ElementData(obj.tag, value, content, attributes, xmlns)
             if 'element_hook' in kwargs:
                 element_data = kwargs['element_hook'](element_data, self, xsd_type)
 
@@ -826,7 +827,7 @@ class XsdElement(XsdComponent, ParticleMixin,
             except (ValueError, TypeError) as err:
                 yield self.validation_error(validation, err, obj, **kwargs)
         elif not level:
-            yield ElementData(obj.tag, value, None, attributes)
+            yield ElementData(obj.tag, value, None, attributes, None)
 
         if content is not None:
             del content
@@ -858,11 +859,10 @@ class XsdElement(XsdComponent, ParticleMixin,
             for identity in self.identities:
                 identities[identity].enabled = False
 
-    def collect_key_fields(self, elem: ElementType, xsd_type: BaseXsdType,
+    def collect_key_fields(self, obj: ElementType, xsd_type: BaseXsdType,
                            validation: str = 'lax', nilled: bool = False,
                            **kwargs: Any) -> Iterator[XMLSchemaValidationError]:
         element_node: Union[ElementNode, LazyElementNode]
-        xsd_fields: Optional[IdentityCounterType]
 
         try:
             identities = kwargs['identities']
@@ -876,14 +876,14 @@ class XsdElement(XsdComponent, ParticleMixin,
         except KeyError:
             namespaces = None
 
-        element_node = resource.get_xpath_node(elem)
+        element_node = resource.get_xpath_node(obj, namespaces)
 
         xsd_element = self if self.ref is None else self.ref
         if xsd_element.type is not xsd_type:
             xsd_element = _copy(xsd_element)
             xsd_element.type = xsd_type
 
-        # Collect field values for identities that refer to this element.
+        # Collect field values for identities that refer to this XSD element.
         for identity in self.selected_by:
             try:
                 counter = identities[identity]
@@ -893,38 +893,31 @@ class XsdElement(XsdComponent, ParticleMixin,
                 if not counter.enabled or not identity.elements:
                     continue
 
-            # Apply selector on Element ancestor for discard false matches
-            root_node = resource.get_xpath_node(counter.elem)
-            context = XPathContext(root_node)
+            if counter.elements is None:
+                # Apply selector on Element ancestor for obtain the selected elements
+                root_node = resource.get_xpath_node(counter.elem)
+                context = XPathContext(root_node)
+                assert identity.selector is not None
+                counter.elements = set(identity.selector.token.select_results(context))
 
-            for item in identity.selector.token.select_results(context):  # type: ignore
-                if item is elem:
-                    break
-            else:
+            if obj not in counter.elements:
                 continue
 
+            if xsd_element in identity.elements:
+                selectors = identity.elements[xsd_element]
+            else:
+                selectors = [FieldValueSelector(f, xsd_element) for f in identity.fields]
+
             try:
-                if xsd_element.type is self.type and xsd_element in identity.elements:
-                    xsd_fields = identity.elements[xsd_element]
-                    if xsd_fields is None:
-                        xsd_fields = identity.get_fields(xsd_element.xpath_node)
-                        identity.elements[xsd_element] = xsd_fields
-                else:
-                    xsd_fields = identity.get_fields(xsd_element.xpath_node)
-
-                if all(x is None for x in xsd_fields):
-                    continue
-
-                decoders = cast(Tuple[XsdAttribute, ...], xsd_fields)
-                fields = identity.get_fields(element_node, namespaces, decoders=decoders)
+                fields = tuple(s.get_value(element_node, namespaces) for s in selectors)
             except (XMLSchemaValueError, XMLSchemaTypeError) as err:
-                yield self.validation_error(validation, err, elem, **kwargs)
+                yield self.validation_error(validation, err, obj, **kwargs)
             else:
                 if any(x is not None for x in fields) or nilled:
                     try:
                         counter.increase(fields)
                     except ValueError as err:
-                        yield self.validation_error(validation, err, elem, **kwargs)
+                        yield self.validation_error(validation, err, obj, **kwargs)
 
     def to_objects(self, obj: ElementType, with_bindings: bool = False, **kwargs: Any) \
             -> DecodeType['dataobjects.DataElement']:
@@ -955,14 +948,15 @@ class XsdElement(XsdComponent, ParticleMixin,
         validation or encoding errors.
         """
         errors: List[Union[str, Exception]] = []
+        result: Any
 
         try:
             converter = kwargs['converter']
         except KeyError:
-            converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+            converter = self._get_converter(obj, kwargs)
         else:
             if not isinstance(converter, XMLSchemaConverter):
-                converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+                converter = self._get_converter(obj, kwargs)
 
         try:
             level = kwargs['level']
@@ -974,15 +968,36 @@ class XsdElement(XsdComponent, ParticleMixin,
         except (ValueError, TypeError) as err:
             yield self.validation_error(validation, err, obj, **kwargs)
             return
-        else:
-            if not level:
-                if not self.is_matching(element_data.tag):
-                    errors.append("data tag does not match XSD element name")
 
-                if 'max_depth' in kwargs and kwargs['max_depth'] == 0:
-                    for e in errors:
-                        yield self.validation_error(validation, e, **kwargs)
-                    return
+        if self.abstract:
+            if self.name == element_data.tag and converter.losslessly:
+                reason = _("can't use an abstract element in an instance")
+                yield self.validation_error(validation, reason, obj, **kwargs)
+            elif self.name not in self.maps.substitution_groups:
+                reason = _("can't use an abstract XSD element for validation "
+                           "unless it's the head of a substitution group")
+                yield self.validation_error(validation, reason, obj, **kwargs)
+            else:
+                for xsd_element in self.iter_substitutes():
+                    if element_data.tag == xsd_element.name:
+                        yield from xsd_element.iter_encode(obj, validation, **kwargs)
+                        return
+                else:
+                    # In some cases the original tag could be missed, so try each
+                    # substitute before generate an error.
+                    for xsd_element in self.iter_substitutes():
+                        for result in xsd_element.iter_encode(obj, validation, **kwargs):
+                            if not isinstance(result, XMLSchemaValidationError):
+                                yield result
+                                return
+                    else:
+                        reason = _("can't use an abstract XSD element for validation")
+                        yield self.validation_error(validation, reason, obj, **kwargs)
+
+        if 'max_depth' in kwargs and kwargs['max_depth'] == 0 and not level:
+            for e in errors:
+                yield self.validation_error(validation, e, **kwargs)
+            return
 
         text = None
         children = element_data.content
@@ -1013,7 +1028,6 @@ class XsdElement(XsdComponent, ParticleMixin,
                             del element_data.attributes[k]
 
         attribute_group = self.get_attributes(xsd_type)
-        result: Any
         for result in attribute_group.iter_encode(element_data.attributes, validation, **kwargs):
             if isinstance(result, XMLSchemaValidationError):
                 errors.append(result)
@@ -1024,7 +1038,7 @@ class XsdElement(XsdComponent, ParticleMixin,
             xsi_nil = element_data.attributes[XSI_NIL].strip()
             if not self.nillable:
                 errors.append("element is not nillable.")
-            elif xsi_nil not in {'0', '1', 'true', 'false'}:
+            elif xsi_nil not in ('0', '1', 'true', 'false'):
                 errors.append("xsi:nil attribute must has a boolean value.")
             elif xsi_nil in ('0', 'false'):
                 pass
@@ -1052,7 +1066,7 @@ class XsdElement(XsdComponent, ParticleMixin,
 
             elif self.fixed is not None:
                 text = self.fixed
-            elif self.default is not None and kwargs.get('use_defaults'):
+            elif self.default is not None and kwargs.get('use_defaults', True):
                 text = self.default
 
         elif xsd_type.has_simple_content():
@@ -1066,7 +1080,7 @@ class XsdElement(XsdComponent, ParticleMixin,
 
             elif self.fixed is not None:
                 text = self.fixed
-            elif self.default is not None and kwargs.get('use_defaults'):
+            elif self.default is not None and kwargs.get('use_defaults', True):
                 text = self.default
 
         else:
@@ -1092,7 +1106,6 @@ class XsdElement(XsdComponent, ParticleMixin,
             name = f'{{{default_namespace}}}{name}'
 
             # Workaround for backward compatibility of XPath selectors on schemas.
-            # It could be removed in v3.0.
             if not self.qualified and default_namespace == self.target_namespace:
                 return (name == self.qualified_name or
                         any(name == e.qualified_name for e in self.iter_substitutes()))
@@ -1114,6 +1127,22 @@ class XsdElement(XsdComponent, ParticleMixin,
                     return xsd_element
         return None
 
+    def match_child(self, name: str) -> Optional['XsdElement']:
+        xsd_group = self.type.model_group
+        if xsd_group is None:
+            # fallback to xs:anyType encoder for matching extra content
+            xsd_group = self.any_type.model_group
+            assert xsd_group is not None
+
+        for xsd_child in xsd_group.iter_elements():
+            matched_element = xsd_child.match(name, resolve=True)
+            if isinstance(matched_element, XsdElement):
+                return matched_element
+        else:
+            if name in self.maps.elements and xsd_group.open_content_mode != 'none':
+                return self.maps.lookup_element(name)
+            return None
+
     def is_restriction(self, other: ModelParticleType, check_occurs: bool = True) -> bool:
         e: ModelParticleType
 
@@ -1413,6 +1442,49 @@ class Xsd11Element(XsdElement):
             warnings.warn(msg.format(e1, e2), XMLSchemaTypeTableWarning, stacklevel=3)
         return True
 
+    def check_dynamic_context(self, elem: ElementType,
+                              validation: str,
+                              options: Dict[str, Any]) -> Iterator[XMLSchemaValidationError]:
+        try:
+            source = options['source']
+        except KeyError:
+            return
+
+        for ns, url in etree_iter_location_hints(elem):
+            base_url = source.base_url
+            url = normalize_url(url, base_url)
+            if any(url == schema.url for schema in self.maps.iter_schemas()):
+                continue
+
+            try:
+                if ns in self.maps.namespaces:
+                    schema = self.maps.namespaces[ns][0]
+                    schema.include_schema(url)
+                    schema.clear()
+                    schema.build()
+                else:
+                    schema = self.schema
+                    schema.import_schema(ns, url, base_url, build=True)
+
+                def stop_validation(e: ElementType, _xsd_element: XsdElement) -> bool:
+                    if e is elem:
+                        raise XMLSchemaStopValidation()
+                    return False
+
+                errors = list(schema.iter_errors(source, validation_hook=stop_validation))
+                if len(options['errors']) != len(errors) or \
+                        any(e1.elem is not e2.elem for e1, e2 in zip(options['errors'], errors)):
+                    reason = _(f"adding schema at {url} change the "
+                               f"assessment outcome of previous items")
+                    yield self.validation_error(validation, reason, elem, **options)
+
+            except (XMLSchemaValidationError, ParseError) as err:
+                yield self.validation_error(validation, err, elem, **options)
+            except XMLSchemaParseError as err:
+                yield self.validation_error(validation, err.message, elem, **options)
+            except OSError:
+                continue
+
 
 class XsdAlternative(XsdComponent):
     """
@@ -1434,7 +1506,7 @@ class XsdAlternative(XsdComponent):
     _ADMITTED_TAGS = {XSD_ALTERNATIVE}
 
     def __init__(self, elem: ElementType, schema: SchemaType, parent: XsdElement) -> None:
-        super(XsdAlternative, self).__init__(elem, schema, parent)
+        super().__init__(elem, schema, parent)
 
     def __repr__(self) -> str:
         return '%s(type=%r, test=%r)' % (
