@@ -12,8 +12,9 @@ This module contains classes for XML Schema model groups.
 """
 import warnings
 from collections.abc import MutableMapping
-from typing import TYPE_CHECKING, overload, Any, Iterable, Iterator, List, \
-    MutableSequence, Optional, Tuple, Union
+from copy import copy as _copy
+from typing import TYPE_CHECKING, cast, overload, Any, Iterable, Iterator, \
+    List, MutableSequence, Optional, Tuple, Union
 from xml.etree import ElementTree
 
 from .. import limits
@@ -21,7 +22,8 @@ from ..exceptions import XMLSchemaValueError
 from ..names import XSD_GROUP, XSD_SEQUENCE, XSD_ALL, XSD_CHOICE, XSD_ELEMENT, \
     XSD_ANY, XSI_TYPE, XSD_ANY_TYPE, XSD_ANNOTATION
 from ..aliases import ElementType, NamespacesType, SchemaType, IterDecodeType, \
-    IterEncodeType, ModelParticleType, SchemaElementType, ComponentClassType
+    IterEncodeType, ModelParticleType, SchemaElementType, ComponentClassType, \
+    OccursCounterType
 from ..translation import gettext as _
 from ..helpers import get_qname, local_name, raw_xml_encode
 from ..converters import ElementData
@@ -32,8 +34,9 @@ from .exceptions import XMLSchemaModelError, XMLSchemaModelDepthError, \
 from .xsdbase import ValidationMixin, XsdComponent, XsdType
 from .particles import ParticleMixin, OccursCalculator
 from .elements import XsdElement, XsdAlternative
-from .wildcards import XsdAnyElement, Xsd11AnyElement
-from .models import ModelVisitor, distinguishable_paths
+from .wildcards import XsdAnyElement, Xsd11AnyElement, XsdOpenContent
+from .models import ModelVisitor, InterleavedModelVisitor, SuffixedModelVisitor, \
+    iter_unordered_content, iter_collapsed_content
 
 if TYPE_CHECKING:
     from .complex_types import XsdComplexType
@@ -93,12 +96,12 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
     parent: Optional[Union['XsdComplexType', 'XsdGroup']]
     model: str
     mixed: bool = False
-    ref: Optional['XsdGroup']
+    ref: Optional['XsdGroup']  # Not None if the instance is a ref to a global group
+    content: List[ModelParticleType]  # Direct access to children also from a ref group
     restriction: Optional['XsdGroup'] = None
 
     # For XSD 1.1 openContent processing
-    interleave: Optional[Xsd11AnyElement] = None  # if openContent with mode='interleave'
-    suffix: Optional[Xsd11AnyElement] = None  # if openContent with mode='suffix'/'interleave'
+    open_content: Optional[XsdOpenContent] = None
 
     _ADMITTED_TAGS = {XSD_GROUP, XSD_SEQUENCE, XSD_ALL, XSD_CHOICE}
 
@@ -107,9 +110,11 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                  parent: Optional[Union['XsdComplexType', 'XsdGroup']] = None) -> None:
 
         self._group: List[ModelParticleType] = []
+        self.content = self._group
+        self.oid = (self,)
         if parent is not None and parent.mixed:
             self.mixed = parent.mixed
-        super(XsdGroup, self).__init__(elem, schema, parent)
+        super().__init__(elem, schema, parent)
 
     def __repr__(self) -> str:
         if self.name is None:
@@ -191,17 +196,35 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         else:
             return True
 
+    @property
+    def open_content_mode(self) -> str:
+        return 'none' if self.open_content is None else self.open_content.mode
+
     @property
     def effective_min_occurs(self) -> int:
         if not self.min_occurs or not self:
             return 0
+
+        effective_items: List[Any]
+        min_occurs: int
+        effective_items = [e for e in self.iter_model() if e.effective_max_occurs != 0]
+        if not effective_items:
+            return 0
         elif self.model == 'choice':
-            if any(not e.effective_min_occurs for e in self.iter_model()):
-                return 0
-        else:
-            if all(not e.effective_min_occurs for e in self.iter_model()):
-                return 0
-        return self.min_occurs
+            min_occurs = min(e.effective_min_occurs for e in effective_items)
+            return self.min_occurs * min_occurs
+        elif self.model == 'all':
+            min_occurs = max(e.effective_min_occurs for e in effective_items)
+            return min_occurs
+
+        not_emptiable_items = [e for e in effective_items if e.effective_min_occurs]
+        if not not_emptiable_items:
+            return 0
+        elif len(not_emptiable_items) > 1:
+            return self.min_occurs
+
+        min_occurs = not_emptiable_items[0].effective_min_occurs
+        return self.min_occurs * min_occurs
 
     @property
     def effective_max_occurs(self) -> Optional[int]:
@@ -209,35 +232,41 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
             return 0
 
         effective_items: List[Any]
-        value: int
+        max_occurs: int
 
-        effective_items = [e for e in self.iter_model() if e.effective_max_occurs != 0]
+        model_items = [(e, e.effective_max_occurs) for e in self.iter_model()]
+        effective_items = [x for x in model_items if x[1] != 0]
         if not effective_items:
             return 0
         elif self.max_occurs is None:
             return None
         elif self.model == 'choice':
-            try:
-                value = max(e.effective_max_occurs for e in effective_items)
-            except TypeError:
+            if any(x[1] is None for x in effective_items):
                 return None
             else:
-                return self.max_occurs * value
+                max_occurs = max(x[1] for x in effective_items)
+                return self.max_occurs * max_occurs
 
-        not_emptiable_items = [e for e in effective_items if e.effective_min_occurs]
+        not_emptiable_items = [x for x in effective_items if x[0].effective_min_occurs]
         if not not_emptiable_items:
-            try:
-                value = max(e.effective_max_occurs for e in effective_items)
-            except TypeError:
+            if any(x[1] is None for x in effective_items):
                 return None
             else:
-                return self.max_occurs * value
+                max_occurs = max(x[1] for x in effective_items)
+                return self.max_occurs * max_occurs
 
         elif len(not_emptiable_items) > 1:
-            return self.max_occurs
-
-        value = not_emptiable_items[0].effective_max_occurs
-        return None if value is None else self.max_occurs * value
+            if self.model == 'sequence':
+                return self.max_occurs
+            elif all(x[1] is None for x in not_emptiable_items):
+                return None
+            else:
+                max_occurs = min(x[1] for x in not_emptiable_items if x[1] is not None)
+                return max_occurs
+        elif not_emptiable_items[0][1] is None:
+            return None
+        else:
+            return self.max_occurs * cast(int, not_emptiable_items[0][1])
 
     def has_occurs_restriction(
             self, other: Union[ModelParticleType, ParticleMixin, 'OccursCalculator']) -> bool:
@@ -245,7 +274,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         if not self:
             return True
         elif isinstance(other, XsdGroup):
-            return super(XsdGroup, self).has_occurs_restriction(other)
+            return super().has_occurs_restriction(other)
 
         # Group particle compared to element particle
         if self.max_occurs is None or any(e.max_occurs is None for e in self):
@@ -323,8 +352,11 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         while True:
             for item in particles:
                 if isinstance(item, XsdGroup):
+                    if item.max_occurs == 0:
+                        continue
+
                     iterators.append(particles)
-                    particles = iter(item)
+                    particles = iter(item.content)
                     if len(iterators) > limits.MAX_MODEL_DEPTH:
                         raise XMLSchemaModelDepthError(self)
                     break
@@ -336,68 +368,100 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                 except IndexError:
                     return
 
-    def get_subgroups(self, item: ModelParticleType) -> List['XsdGroup']:
+    def get_subgroups(self, particle: ModelParticleType) -> List['XsdGroup']:
         """
         Returns a list of the groups that represent the path to the enclosed particle.
-        Raises an `XMLSchemaModelError` if *item* is not a particle of the model group.
+        Raises an `XMLSchemaModelError` if the argument is not a particle of the model
+        group.
         """
         subgroups: List[Tuple[XsdGroup, Iterator[ModelParticleType]]] = []
-        group, children = self, iter(self)
+        group, children = self, iter(self if self.ref is None else self.ref)
 
         while True:
-            try:
-                child = next(children)
-            except StopIteration:
+            for child in children:
+                if child is particle:
+                    _subgroups = [x[0] for x in subgroups]
+                    _subgroups.append(group)
+                    return _subgroups
+                elif isinstance(child, XsdGroup):
+                    if len(subgroups) > limits.MAX_MODEL_DEPTH:
+                        raise XMLSchemaModelDepthError(self)
+                    subgroups.append((group, children))
+                    group, children = child, iter(child if child.ref is None else child.ref)
+                    break
+            else:
                 try:
                     group, children = subgroups.pop()
                 except IndexError:
                     msg = _('{!r} is not a particle of the model group')
-                    raise XMLSchemaModelError(self, msg.format(item)) from None
-                else:
-                    continue
+                    raise XMLSchemaModelError(self, msg.format(particle)) from None
 
-            if child is item:
-                _subgroups = [x[0] for x in subgroups]
-                _subgroups.append(group)
-                return _subgroups
-            elif isinstance(child, XsdGroup):
-                if len(subgroups) > limits.MAX_MODEL_DEPTH:
-                    raise XMLSchemaModelDepthError(self)
-                subgroups.append((group, children))
-                group, children = child, iter(child)
+    def get_model_visitor(self) -> ModelVisitor:
+        if self.open_content is None or self.open_content.mode == 'none':
+            return ModelVisitor(self)
+        elif self.open_content.mode == 'interleave':
+            return InterleavedModelVisitor(self, self.open_content.any_element)
+        else:
+            return SuffixedModelVisitor(self, self.open_content.any_element)
 
-    def overall_min_occurs(self, item: ModelParticleType) -> int:
-        """Returns the overall min occurs of a particle in the model."""
-        min_occurs = item.min_occurs
+    def overall_min_occurs(self, particle: ModelParticleType) -> int:
+        """
+        Returns the overall min occurs of a particle in the model group.
+        """
+        model = self.get_model_visitor()
+        return model.overall_min_occurs(particle)
 
-        for group in self.get_subgroups(item):
-            if group.model == 'choice' and len(group) > 1:
-                return 0
-            min_occurs *= group.min_occurs
+    def overall_max_occurs(self, particle: ModelParticleType) -> Optional[int]:
+        """
+        Returns the overall max occurs of a particle in the model group.
+        """
+        model = self.get_model_visitor()
+        return model.overall_max_occurs(particle)
 
-        return min_occurs
+    def is_optional(self, particle: ModelParticleType) -> bool:
+        """
+        Returns `True` if the provided particle can be optional in the model group.
+        """
+        return self.overall_min_occurs(particle) == 0
 
-    def overall_max_occurs(self, item: ModelParticleType) -> Optional[int]:
-        """Returns the overall max occurs of a particle in the model."""
-        max_occurs = item.max_occurs
+    def is_missing(self, occurs: OccursCounterType) -> bool:
+        value = occurs[self.oid] or occurs[self]
+        return not self.is_emptiable() if value == 0 else self.min_occurs > value
 
-        for group in self.get_subgroups(item):
-            if max_occurs == 0:
-                return 0
-            elif max_occurs is None:
-                continue
-            elif group.max_occurs is None:
-                max_occurs = None
+    def get_expected(self, occurs: OccursCounterType) -> List[SchemaElementType]:
+        """
+        Returns the expected elements of the current and descendant groups
+        given a counter of occurrences. Returns an empty list if the group
+        reached the maximum number of occurrences.
+        """
+        expected: List[SchemaElementType] = []
+        items: Union['XsdGroup', Iterator[ModelParticleType]]
+
+        if self.is_over(occurs):
+            return expected
+        elif self.model == 'choice':
+            items = self
+        else:
+            items = (p for p in self._group if p.min_occurs > occurs[p])
+
+        for p in items:
+            if isinstance(p, XsdGroup):
+                expected.extend(
+                    e for e in p.iter_elements() if e.min_occurs > occurs[e]
+                )
             else:
-                max_occurs *= group.max_occurs
-
-        return max_occurs
+                expected.append(p)
+                if p.name in p.maps.substitution_groups:
+                    expected.extend(p.maps.substitution_groups[p.name])
+        return expected
 
     def copy(self) -> 'XsdGroup':
         group: XsdGroup = object.__new__(self.__class__)
         group.__dict__.update(self.__dict__)
         group.errors = self.errors[:]
         group._group = self._group[:]
+        if self.ref is None:
+            group.content = group._group
         return group
 
     __copy__ = copy
@@ -435,6 +499,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                         self.parse_error(msg)
                 self._group.append(xsd_group)
                 self.ref = xsd_group
+                self.content = xsd_group._group
             else:
                 # Disallowed circular definition, substitute with any content group.
                 msg = _("Circular definition detected for group %r")
@@ -669,17 +734,13 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         # Same model: declarations must simply preserve order
         other_iterator = iter(other.iter_model())
         for item in self.iter_model():
-            while True:
-                try:
-                    other_item = next(other_iterator)
-                except StopIteration:
-                    return False
+            for other_item in other_iterator:
                 if other_item is item or item.is_restriction(other_item, check_occurs):
                     break
                 elif other.model == 'choice':
                     if item.max_occurs != 0:
                         continue
-                    elif not other_item.is_matching(item.name, self.default_namespace):
+                    elif not other_item.is_matching(item.name):
                         continue
                     elif all(e.max_occurs == 0 for e in self.iter_model()):
                         return False
@@ -687,6 +748,8 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                         break
                 elif not other_item.is_emptiable():
                     return False
+            else:
+                return False
 
         if other.model != 'choice':
             for other_item in other_iterator:
@@ -769,78 +832,6 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         else:
             return other_max_occurs >= max_occurs * self.max_occurs
 
-    def check_model(self) -> None:
-        """
-        Checks if the model group is deterministic. Element Declarations Consistent and
-        Unique Particle Attribution constraints are checked.
-        :raises: an `XMLSchemaModelError` at first violated constraint.
-        """
-        def safe_iter_path() -> Iterator[SchemaElementType]:
-            iterators: List[Iterator[ModelParticleType]] = []
-            particles = iter(self)
-
-            while True:
-                for item in particles:
-                    if isinstance(item, XsdGroup):
-                        current_path.append(item)
-                        iterators.append(particles)
-                        particles = iter(item)
-                        if len(iterators) > limits.MAX_MODEL_DEPTH:
-                            raise XMLSchemaModelDepthError(self)
-                        break
-                    else:
-                        yield item
-                else:
-                    try:
-                        current_path.pop()
-                        particles = iterators.pop()
-                    except IndexError:
-                        return
-
-        paths: Any = {}
-        current_path: List[ModelParticleType] = [self]
-        try:
-            any_element = self.parent.open_content.any_element  # type: ignore[union-attr]
-        except AttributeError:
-            any_element = None
-
-        for e in safe_iter_path():
-
-            previous_path: List[ModelParticleType]
-            for pe, previous_path in paths.values():
-                # EDC check
-                if not e.is_consistent(pe) or any_element and not any_element.is_consistent(pe):
-                    msg = _("Element Declarations Consistent violation between {0!r} and {1!r}"
-                            ": match the same name but with different types").format(e, pe)
-                    raise XMLSchemaModelError(self, msg)
-
-                # UPA check
-                if pe is e or not pe.is_overlap(e):
-                    continue
-                elif pe.parent is e.parent:
-                    if pe.parent.model in {'all', 'choice'}:
-                        if isinstance(pe, Xsd11AnyElement) and not isinstance(e, XsdAnyElement):
-                            pe.add_precedence(e, self)
-                        elif isinstance(e, Xsd11AnyElement) and not isinstance(pe, XsdAnyElement):
-                            e.add_precedence(pe, self)
-                        else:
-                            msg = _("{0!r} and {1!r} overlap and are in the same {2!r} group")
-                            raise XMLSchemaModelError(self, msg.format(pe, e, pe.parent.model))
-                    elif pe.is_univocal():
-                        continue
-
-                if distinguishable_paths(previous_path + [pe], current_path + [e]):
-                    continue
-                elif isinstance(pe, Xsd11AnyElement) and not isinstance(e, XsdAnyElement):
-                    pe.add_precedence(e, self)
-                elif isinstance(e, Xsd11AnyElement) and not isinstance(pe, XsdAnyElement):
-                    e.add_precedence(pe, self)
-                else:
-                    msg = _("Unique Particle Attribution violation between {0!r} and {1!r}")
-                    raise XMLSchemaModelError(self, msg.format(pe, e))
-
-            paths[e.name] = e, current_path[:]
-
     def check_dynamic_context(self, elem: ElementType,
                               xsd_element: SchemaElementType,
                               model_element: SchemaElementType,
@@ -860,6 +851,10 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
             try:
                 xsd_element = self.maps.lookup_element(elem.tag)
             except LookupError:
+                if self.schema.meta_schema is None:
+                    # Meta-schema groups ignore xsi:type (issue #350)
+                    return
+
                 try:
                     type_name = elem.attrib[XSI_TYPE].strip()
                 except KeyError:
@@ -880,7 +875,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                     )
 
         else:
-            if XSI_TYPE not in elem.attrib:
+            if XSI_TYPE not in elem.attrib or self.schema.meta_schema is None:
                 xsd_type = xsd_element.type
             else:
                 alternatives = xsd_element.alternatives
@@ -902,7 +897,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                                    "head element").format(xsd_element, derivation)
                         raise XMLSchemaValidationError(self, elem, reason)
 
-            if XSI_TYPE not in elem.attrib:
+            if XSI_TYPE not in elem.attrib or self.schema.meta_schema is None:
                 return
 
         # If it's a restriction the context is the base_type's group
@@ -933,14 +928,13 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                         "{0!r} and {1!r}.").format(self, xsd_element)
                 warnings.warn(msg, XMLSchemaTypeTableWarning, stacklevel=3)
 
-    def match_element(self, name: str, default_namespace: Optional[str] = None) \
-            -> Optional[SchemaElementType]:
+    def match_element(self, name: str) -> Optional[SchemaElementType]:
         """
         Try a model-less match of a child element. Returns the
         matched element, or `None` if there is no match.
         """
         for xsd_element in self.iter_elements():
-            if xsd_element.is_matching(name, default_namespace, group=self):
+            if xsd_element.is_matching(name, group=self):
                 return xsd_element
         return None
 
@@ -981,53 +975,45 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                 result_list.append((cdata_index, text, None))
                 cdata_index += 1
 
-        level = kwargs['level'] = kwargs.pop('level', 0) + 1
+        try:
+            level = kwargs['level'] = kwargs['level'] + 1
+        except KeyError:
+            level = kwargs['level'] = 1
+
         over_max_depth = 'max_depth' in kwargs and kwargs['max_depth'] <= level
         if level > limits.MAX_XML_DEPTH:
             reason = _("XML data depth exceeded (MAX_XML_DEPTH=%r)") % limits.MAX_XML_DEPTH
             self.validation_error('strict', reason, obj, **kwargs)
 
         try:
-            namespaces = kwargs['namespaces']
+            converter = kwargs['converter']
         except KeyError:
-            namespaces = default_namespace = None
-        else:
-            try:
-                default_namespace = namespaces.get('')
-            except AttributeError:
-                default_namespace = None
+            converter = self._get_converter(obj, kwargs)
 
         errors: List[Tuple[int, ModelParticleType, int, Optional[List[SchemaElementType]]]]
         xsd_element: Optional[SchemaElementType]
         expected: Optional[List[SchemaElementType]]
 
-        model = ModelVisitor(self)
         errors = []
         broken_model = False
+        namespaces = converter.namespaces
+        model = self.get_model_visitor()
 
         for index, child in enumerate(obj):
             if callable(child.tag):
-                continue  # child is a <class 'lxml.etree._Comment'>
+                continue  # child is a comment or PI
+
+            converter.set_context(child, level)
+            name = converter.map_qname(child.tag)
 
             while model.element is not None:
-                if model.element.max_occurs == 0:
-                    xsd_element = None
-                else:
-                    xsd_element = model.element.match(
-                        child.tag, default_namespace, group=self, occurs=model.occurs
-                    )
-
+                xsd_element = model.match_element(child.tag)
                 if xsd_element is None:
-                    if self.interleave is not None and self.interleave.is_matching(
-                            child.tag, default_namespace, self, model.occurs):
-                        xsd_element = self.interleave
-                        break
-
                     for particle, occurs, expected in model.advance(False):
                         errors.append((index, particle, occurs, expected))
                         model.clear()
                         broken_model = True  # the model is broken, continues with raw decoding.
-                        xsd_element = self.match_element(child.tag, default_namespace)
+                        xsd_element = self.match_element(child.tag)
                         break
                     else:
                         continue
@@ -1042,40 +1028,30 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                     errors.append((index, particle, occurs, expected))
                 break
             else:
-                if self.suffix is not None and \
-                        self.suffix.is_matching(child.tag, default_namespace, self):
-                    xsd_element = self.suffix
-                else:
-                    xsd_element = self.match_element(child.tag, default_namespace)
-                    if xsd_element is None:
-                        errors.append((index, self, 0, None))
-                        broken_model = True
-                    elif not broken_model:
-                        errors.append((index, xsd_element, 0, []))
-                        broken_model = True
+                xsd_element = self.match_element(child.tag)
+                if xsd_element is None:
+                    errors.append((index, self, 0, None))
+                    broken_model = True
+                elif not broken_model:
+                    errors.append((index, xsd_element, 0, []))
+                    broken_model = True
 
             if xsd_element is None:
-                if kwargs.get('keep_unknown') and 'converter' in kwargs:
+                if kwargs.get('keep_unknown'):
                     for result in self.any_type.iter_decode(child, validation, **kwargs):
-                        result_list.append((child.tag, result, None))
-                continue
-            elif 'converter' not in kwargs:
-                # Validation-only mode: do not append results
-                for result in xsd_element.iter_decode(child, validation, **kwargs):
-                    if isinstance(result, XMLSchemaValidationError):
-                        yield result
+                        result_list.append((name, result, None))
                 continue
             elif over_max_depth:
                 if 'depth_filler' in kwargs:
                     func = kwargs['depth_filler']
-                    result_list.append((child.tag, func(xsd_element), xsd_element))
+                    result_list.append((name, func(xsd_element), xsd_element))
                 continue
 
             for result in xsd_element.iter_decode(child, validation, **kwargs):
                 if isinstance(result, XMLSchemaValidationError):
                     yield result
                 else:
-                    result_list.append((child.tag, result, xsd_element))
+                    result_list.append((name, result, xsd_element))
 
             if cdata_index and child.tail is not None:
                 tail = str(child.tail.strip())
@@ -1091,12 +1067,22 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
             index = len(obj)
             for particle, occurs, expected in model.stop():
                 errors.append((index, particle, occurs, expected))
+                break
 
         if errors:
             source = kwargs.get('source')
+            namespaces = converter.namespaces
+
             for index, particle, occurs, expected in errors:
                 error = XMLSchemaChildrenValidationError(
-                    self, obj, index, particle, occurs, expected, source, namespaces
+                    validator=self,
+                    elem=obj,
+                    index=index,
+                    particle=particle,
+                    occurs=occurs,
+                    expected=expected,
+                    source=source,
+                    namespaces=namespaces,
                 )
                 if validation == 'strict':
                     raise error
@@ -1115,35 +1101,29 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         :return: yields a couple with the text of the Element and a list of child \
         elements, eventually preceded by a sequence of validation errors.
         """
-        level = kwargs['level'] = kwargs.get('level', 0) + 1
         errors = []
         text = raw_xml_encode(obj.text)
         children: List[ElementType] = []
-        try:
-            indent = kwargs['indent']
-        except KeyError:
-            indent = 4
-
-        padding = '\n' + ' ' * indent * level
 
         try:
-            converter = kwargs['converter']
+            level = kwargs['level'] = kwargs['level'] + 1
         except KeyError:
-            converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+            level = kwargs['level'] = 1
 
+        converter = kwargs['converter']
+        padding = '\n' + ' ' * converter.indent * level
         default_namespace = converter.get('')
-        model = ModelVisitor(self)
+
         index = cdata_index = 0
         wrong_content_type = False
         over_max_depth = 'max_depth' in kwargs and kwargs['max_depth'] <= level
+        model = self.get_model_visitor()
 
         content: Iterable[Any]
         if not obj.content:
             content = []
         elif isinstance(obj.content, MutableMapping) or kwargs.get('unordered'):
-            content = ModelVisitor(self).iter_unordered_content(
-                obj.content, default_namespace
-            )
+            content = iter_unordered_content(obj.content, self)
         elif not isinstance(obj.content, MutableSequence):
             wrong_content_type = True
             content = []
@@ -1156,9 +1136,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         elif converter.losslessly:
             content = obj.content
         else:
-            content = ModelVisitor(self).iter_collapsed_content(
-                obj.content, default_namespace
-            )
+            content = iter_collapsed_content(obj.content, self)
 
         for index, (name, value) in enumerate(content):
             if isinstance(name, int):
@@ -1172,47 +1150,33 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                 continue
 
             xsd_element: Optional[SchemaElementType]
-            if self.interleave and self.interleave.is_matching(name, default_namespace, group=self):
-                xsd_element = self.interleave
-                value = get_qname(default_namespace, name), value
-            else:
-                while model.element is not None:
-                    if model.element.max_occurs == 0:
-                        xsd_element = None
-                    else:
-                        xsd_element = model.element.match(
-                            name, default_namespace, group=self, occurs=model.occurs
-                        )
-
-                    if xsd_element is None:
-                        for particle, occurs, expected in model.advance():
-                            errors.append((index - cdata_index, particle, occurs, expected))
-                        continue
-                    elif isinstance(xsd_element, XsdAnyElement):
-                        value = get_qname(default_namespace, name), value
-
-                    for particle, occurs, expected in model.advance(True):
+            while model.element is not None:
+                xsd_element = model.match_element(name)
+                if xsd_element is None:
+                    for particle, occurs, expected in model.advance():
                         errors.append((index - cdata_index, particle, occurs, expected))
-                    break
-                else:
-                    if self.suffix and self.suffix.is_matching(name, default_namespace, group=self):
-                        xsd_element = self.suffix
-                        value = get_qname(default_namespace, name), value
+                    continue
+                elif isinstance(xsd_element, XsdAnyElement):
+                    value = get_qname(default_namespace, name), value
+
+                for particle, occurs, expected in model.advance(True):
+                    errors.append((index - cdata_index, particle, occurs, expected))
+                break
+            else:
+                errors.append((index - cdata_index, self, 0, []))
+                xsd_element = self.match_element(name)
+                if isinstance(xsd_element, XsdAnyElement):
+                    value = get_qname(default_namespace, name), value
+                elif xsd_element is None:
+                    if name.startswith('{') or ':' not in name:
+                        reason = _('{!r} does not match any declared element '
+                                   'of the model group').format(name)
                     else:
-                        errors.append((index - cdata_index, self, 0, []))
-                        xsd_element = self.match_element(name, default_namespace)
-                        if isinstance(xsd_element, XsdAnyElement):
-                            value = get_qname(default_namespace, name), value
-                        elif xsd_element is None:
-                            if name.startswith('{') or ':' not in name:
-                                reason = _('{!r} does not match any declared element '
-                                           'of the model group').format(name)
-                            else:
-                                reason = _('{0} has an unknown prefix {1!r}').format(
-                                    name, name.split(':')[0]
-                                )
-                            yield self.validation_error(validation, reason, value, **kwargs)
-                            continue
+                        reason = _('{0} has an unknown prefix {1!r}').format(
+                            name, name.split(':')[0]
+                        )
+                    yield self.validation_error(validation, reason, value, **kwargs)
+                    continue
 
             if over_max_depth:
                 continue
@@ -1226,12 +1190,15 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         if model.element is not None:
             for particle, occurs, expected in model.stop():
                 errors.append((index - cdata_index + 1, particle, occurs, expected))
+                break
 
         if children:
             if children[-1].tail is None:
-                children[-1].tail = padding[:-indent] or '\n'
+                children[-1].tail = padding[:-converter.indent] or '\n'
             else:
-                children[-1].tail = children[-1].tail.strip() + (padding[:-indent] or '\n')
+                children[-1].tail = children[-1].tail.strip() + (
+                        padding[:-converter.indent] or '\n'
+                )
 
         cdata_not_allowed = not self.mixed and text and text.strip() and self and \
             (len(self) > 1 or not isinstance(self[0], XsdAnyElement))
@@ -1239,6 +1206,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
         if errors or cdata_not_allowed or wrong_content_type:
             attrib = {k: raw_xml_encode(v) for k, v in obj.attributes.items()}
             elem = converter.etree_element(obj.tag, text, children, attrib)
+            namespaces = converter.namespaces
 
             if wrong_content_type:
                 reason = _("wrong content type {!r}").format(type(obj.content))
@@ -1256,7 +1224,7 @@ class XsdGroup(XsdComponent, MutableSequence[ModelParticleType],
                     particle=particle,
                     occurs=occurs,
                     expected=expected,
-                    namespaces=converter.namespaces,
+                    namespaces=namespaces,
                 )
                 if validation == 'strict':
                     raise error
@@ -1272,7 +1240,7 @@ class Xsd11Group(XsdGroup):
     Class for XSD 1.1 *model group* definitions.
 
     .. The XSD 1.1 model groups differ from XSD 1.0 groups for the 'all' model,
-    .. that can contains also other groups.
+       that can contains also other groups.
     ..  <all
           id = ID
           maxOccurs = (0 | 1) : 1
@@ -1360,6 +1328,26 @@ class Xsd11Group(XsdGroup):
         else:  # other.model == 'choice':
             return self.is_choice_restriction(other)
 
+    def has_occurs_restriction(
+            self, other: Union[ModelParticleType, ParticleMixin, 'OccursCalculator']) -> bool:
+        if not isinstance(other, XsdGroup):
+            return super().has_occurs_restriction(other)
+        elif not self:
+            return True
+        elif self.effective_min_occurs < other.effective_min_occurs:
+            return False
+
+        effective_max_occurs = self.effective_max_occurs
+        if effective_max_occurs == 0:
+            return True
+        elif effective_max_occurs is None:
+            return other.effective_max_occurs is None
+
+        try:
+            return effective_max_occurs <= other.effective_max_occurs  # type: ignore[operator]
+        except TypeError:
+            return True
+
     def is_sequence_restriction(self, other: XsdGroup) -> bool:
         if not self.has_occurs_restriction(other):
             return False
@@ -1385,13 +1373,23 @@ class Xsd11Group(XsdGroup):
         for other_item in other.iter_model():
             if item is not None and item.is_restriction(other_item, check_occurs):
                 item = next(item_iterator, None)
+            elif not other_item.is_emptiable():
+                break
+        else:
+            if item is None:
+                return True
+
+        # Restriction check failed again: try checking other items against self
+        other_items = other.iter_model()
+        for other_item in other_items:
+            if self.is_restriction(other_item, check_occurs):
+                return all(x.is_emptiable() for x in other_items)
             elif not other_item.is_emptiable():
                 return False
-        return item is None
+        else:
+            return False
 
     def is_all_restriction(self, other: XsdGroup) -> bool:
-        if not self.has_occurs_restriction(other):
-            return False
         restriction_items = [x for x in self.iter_model()]
 
         base_items = [x for x in other.iter_model()]
@@ -1407,7 +1405,7 @@ class Xsd11Group(XsdGroup):
                         w2.extended = True
                         break
                 else:
-                    wildcards.append(w1.copy())
+                    wildcards.append(_copy(w1))
 
         base_items.extend(w for w in wildcards if hasattr(w, 'extended'))
 
@@ -1532,7 +1530,7 @@ class Xsd11Group(XsdGroup):
                     break
                 elif item.max_occurs != 0:
                     continue
-                elif not other_item.is_matching(item.name, self.default_namespace):
+                elif not other_item.is_matching(item.name):
                     continue
                 elif has_not_empty_item:
                     break
