@@ -25,9 +25,9 @@ from ..names import XSD_NAMESPACE, XSD_ANY_TYPE, XSD_SIMPLE_TYPE, XSD_PATTERN, \
     XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_WHITE_SPACE, XSD_ENUMERATION, \
     XSD_LIST, XSD_ANY_SIMPLE_TYPE, XSD_UNION, XSD_RESTRICTION, XSD_ANNOTATION, \
     XSD_ASSERTION, XSD_ID, XSD_IDREF, XSD_FRACTION_DIGITS, XSD_TOTAL_DIGITS, \
-    XSD_EXPLICIT_TIMEZONE, XSD_ERROR, XSD_ASSERT, XSD_QNAME
+    XSD_EXPLICIT_TIMEZONE, XSD_ERROR, XSD_ASSERT, XSD_QNAME, XSD_NOTATION
 from ..translation import gettext as _
-from ..helpers import local_name
+from ..helpers import local_name, get_extended_qname
 
 from .exceptions import XMLSchemaValidationError, XMLSchemaEncodeError, \
     XMLSchemaDecodeError, XMLSchemaParseError
@@ -38,7 +38,7 @@ from .facets import XsdFacet, XsdWhiteSpaceFacet, XsdPatternFacets, \
     XSD_11_LIST_FACETS, XSD_10_UNION_FACETS, XSD_11_UNION_FACETS, MULTIPLE_FACETS
 
 FacetsValueType = Union[XsdFacet, Callable[[Any], None], List[XsdAssertionFacet]]
-PythonTypeClasses = Type[Any]
+PythonTypeClasses = Union[Type[Any], Tuple[Type[Any]]]
 
 
 class XsdSimpleType(XsdType, ValidationMixin[Union[str, bytes], DecodedValueType]):
@@ -67,8 +67,10 @@ class XsdSimpleType(XsdType, ValidationMixin[Union[str, bytes], DecodedValueType
     allow_empty = True
     facets: Dict[Optional[str], FacetsValueType]
 
-    python_type: PythonTypeClasses
-    instance_types: Union[PythonTypeClasses, Tuple[PythonTypeClasses]]
+    python_type: Type[Any]
+    instance_types: PythonTypeClasses
+    to_python: Union[Type[Any], Callable[[Any], Any]]
+    from_python: Union[Type[str], Callable[[Any], Any]]
 
     # Unicode string as default datatype for XSD simple types
     python_type = instance_types = to_python = from_python = str
@@ -79,12 +81,12 @@ class XsdSimpleType(XsdType, ValidationMixin[Union[str, bytes], DecodedValueType
                  name: Optional[str] = None,
                  facets: Optional[Dict[Optional[str], FacetsValueType]] = None) -> None:
 
-        super(XsdSimpleType, self).__init__(elem, schema, parent, name)
+        super().__init__(elem, schema, parent, name)
         if not hasattr(self, 'facets'):
             self.facets = facets if facets is not None else {}
 
     def __setattr__(self, name: str, value: Any) -> None:
-        super(XsdSimpleType, self).__setattr__(name, value)
+        super().__setattr__(name, value)
         if name == 'facets':
             if not isinstance(self, XsdAtomicBuiltin):
                 self._parse_facets(value)
@@ -345,6 +347,15 @@ class XsdSimpleType(XsdType, ValidationMixin[Union[str, bytes], DecodedValueType
     def content_type_label(self) -> str:
         return 'empty' if self.max_length == 0 else 'simple'
 
+    @property
+    def root_type(self) -> BaseXsdType:
+        if self.base_type is None:
+            return self
+        elif isinstance(self.base_type, XsdAtomic):
+            return self.base_type.primitive_type
+        else:
+            return self.base_type.root_type
+
     @property
     def sequence_type(self) -> str:
         if self.is_empty():
@@ -502,7 +513,7 @@ class XsdAtomic(XsdSimpleType):
             self.primitive_type = self
         else:
             self.base_type = base_type
-        super(XsdAtomic, self).__init__(elem, schema, parent, name, facets)
+        super().__init__(elem, schema, parent, name, facets)
 
     def __repr__(self) -> str:
         if self.name is None:
@@ -513,7 +524,7 @@ class XsdAtomic(XsdSimpleType):
             return '%s(name=%r)' % (self.__class__.__name__, self.prefixed_name)
 
     def __setattr__(self, name: str, value: Any) -> None:
-        super(XsdAtomic, self).__setattr__(name, value)
+        super().__setattr__(name, value)
         if name == 'base_type':
             if not hasattr(self, 'white_space'):
                 try:
@@ -568,12 +579,12 @@ class XsdAtomicBuiltin(XsdAtomic):
     def __init__(self, elem: ElementType,
                  schema: SchemaType,
                  name: str,
-                 python_type: Type[Any],
+                 python_type: PythonTypeClasses,
                  base_type: Optional['XsdAtomicBuiltin'] = None,
                  admitted_facets: Optional[Set[str]] = None,
                  facets: Optional[Dict[Optional[str], FacetsValueType]] = None,
-                 to_python: Any = None,
-                 from_python: Any = None) -> None:
+                 to_python: Optional[Callable[[Any], Any]] = None,
+                 from_python: Optional[Callable[[Any], Any]] = None) -> None:
         """
         :param name: the XSD type's qualified name.
         :param python_type: the correspondent Python's type. If a tuple of types \
@@ -588,15 +599,16 @@ class XsdAtomicBuiltin(XsdAtomic):
             self.instance_types, python_type = python_type, python_type[0]
         else:
             self.instance_types = python_type
-        if not callable(python_type):
-            raise XMLSchemaTypeError("%r object is not callable" % python_type.__class__)
+
+        if not isinstance(python_type, type):
+            raise XMLSchemaTypeError(f"{python_type!r} object is not a type")
 
         if base_type is None and not admitted_facets and name != XSD_ERROR:
             raise XMLSchemaValueError("argument 'admitted_facets' must be "
                                       "a not empty set of a primitive type")
         self._admitted_facets = admitted_facets
 
-        super(XsdAtomicBuiltin, self).__init__(elem, schema, None, name, facets, base_type)
+        super().__init__(elem, schema, None, name, facets, base_type)
         self.python_type = python_type
         self.to_python = to_python if to_python is not None else python_type
         self.from_python = from_python if from_python is not None else str
@@ -614,7 +626,8 @@ class XsdAtomicBuiltin(XsdAtomic):
             obj = self.normalize(obj)
         elif obj is not None and not isinstance(obj, self.instance_types):
             reason = _("value is not an instance of {!r}").format(self.instance_types)
-            yield XMLSchemaDecodeError(self, obj, self.to_python, reason)
+            error = XMLSchemaDecodeError(self, obj, self.to_python, reason)
+            yield self.validation_error(validation, error, **kwargs)
 
         if validation == 'skip':
             try:
@@ -632,13 +645,14 @@ class XsdAtomicBuiltin(XsdAtomic):
         try:
             result = self.to_python(obj)
         except (ValueError, DecimalException) as err:
-            yield XMLSchemaDecodeError(self, obj, self.to_python, reason=str(err))
+            error = XMLSchemaDecodeError(self, obj, self.to_python, reason=str(err))
+            yield self.validation_error(validation, error, **kwargs)
             yield None
             return
         except TypeError:
             # xs:error type (e.g. an XSD 1.1 type alternative used to catch invalid values)
             reason = _("invalid value {!r}").format(obj)
-            yield self.validation_error(validation, error=reason, obj=obj)
+            yield self.validation_error(validation, reason, obj, **kwargs)
             yield None
             return
 
@@ -648,7 +662,7 @@ class XsdAtomicBuiltin(XsdAtomic):
             except XMLSchemaValidationError as err:
                 yield err
 
-        if self.name not in {XSD_QNAME, XSD_IDREF, XSD_ID}:
+        if self.name not in (XSD_QNAME, XSD_IDREF, XSD_ID):
             pass
         elif self.name == XSD_QNAME:
             if ':' in obj:
@@ -663,7 +677,7 @@ class XsdAtomicBuiltin(XsdAtomic):
                         try:
                             if kwargs['source'].namespace != XSD_NAMESPACE:
                                 reason = _("unmapped prefix %r in a QName") % prefix
-                                yield self.validation_error(validation, error=reason, obj=obj)
+                                yield self.validation_error(validation, reason, obj, **kwargs)
                         except KeyError:
                             pass
             else:
@@ -697,7 +711,7 @@ class XsdAtomicBuiltin(XsdAtomic):
                         id_map[obj] = 1
                     else:
                         reason = _("duplicated xs:ID value {!r}").format(obj)
-                        yield self.validation_error(validation, error=reason, obj=obj)
+                        yield self.validation_error(validation, reason, obj, **kwargs)
                 else:
                     if not id_map[obj]:
                         id_map[obj] = 1
@@ -709,7 +723,7 @@ class XsdAtomicBuiltin(XsdAtomic):
 
                     elif obj not in id_list or self.xsd_version == '1.0':
                         reason = _("duplicated xs:ID value {!r}").format(obj)
-                        yield self.validation_error(validation, error=reason, obj=obj)
+                        yield self.validation_error(validation, reason, obj, **kwargs)
 
         yield result
 
@@ -729,12 +743,14 @@ class XsdAtomicBuiltin(XsdAtomic):
             types_: Any = self.instance_types
             if types_ is not bool or (isinstance(types_, tuple) and bool in types_):
                 reason = _("boolean value {0!r} requires a {1!r} decoder").format(obj, bool)
-                yield XMLSchemaEncodeError(self, obj, self.from_python, reason)
+                error = XMLSchemaEncodeError(self, obj, self.from_python, reason)
+                yield self.validation_error(validation, error, **kwargs)
                 obj = self.python_type(obj)
 
         elif not isinstance(obj, self.instance_types):
             reason = _("{0!r} is not an instance of {1!r}").format(obj, self.instance_types)
-            yield XMLSchemaEncodeError(self, obj, self.from_python, reason)
+            error = XMLSchemaEncodeError(self, obj, self.from_python, reason)
+            yield self.validation_error(validation, error, **kwargs)
 
             try:
                 value = self.python_type(obj)
@@ -743,7 +759,8 @@ class XsdAtomicBuiltin(XsdAtomic):
                     raise ValueError()
                 obj = value
             except (ValueError, TypeError) as err:
-                yield XMLSchemaEncodeError(self, obj, self.from_python, reason=str(err))
+                error = XMLSchemaEncodeError(self, obj, self.from_python, reason=str(err))
+                yield self.validation_error(validation, error, **kwargs)
                 yield None
                 return
             else:
@@ -751,7 +768,8 @@ class XsdAtomicBuiltin(XsdAtomic):
                     obj = value
                 else:
                     reason = _("invalid value {!r}").format(obj)
-                    yield XMLSchemaEncodeError(self, obj, self.from_python, reason)
+                    error = XMLSchemaEncodeError(self, obj, self.from_python, reason)
+                    yield self.validation_error(validation, error, **kwargs)
                     yield None
                     return
 
@@ -764,7 +782,8 @@ class XsdAtomicBuiltin(XsdAtomic):
         try:
             text = self.from_python(obj)
         except ValueError as err:
-            yield XMLSchemaEncodeError(self, obj, self.from_python, reason=str(err))
+            error = XMLSchemaEncodeError(self, obj, self.from_python, reason=str(err))
+            yield self.validation_error(validation, error, **kwargs)
             yield None
         else:
             if self.patterns is not None:
@@ -787,7 +806,7 @@ class XsdList(XsdSimpleType):
           Content: (annotation?, simpleType?)
         </list>
     """
-    base_type: XsdSimpleType
+    item_type: XsdSimpleType
     _ADMITTED_TAGS = {XSD_LIST}
     _white_space_elem = ElementTree.Element(
         XSD_WHITE_SPACE, attrib={'value': 'collapse', 'fixed': 'true'}
@@ -800,11 +819,11 @@ class XsdList(XsdSimpleType):
         facets: Optional[Dict[Optional[str], FacetsValueType]] = {
             XSD_WHITE_SPACE: XsdWhiteSpaceFacet(self._white_space_elem, schema, self, self)
         }
-        super(XsdList, self).__init__(elem, schema, parent, name, facets)
+        super().__init__(elem, schema, parent, name, facets)
 
     def __repr__(self) -> str:
         if self.name is None:
-            return '%s(item_type=%r)' % (self.__class__.__name__, self.base_type)
+            return '%s(item_type=%r)' % (self.__class__.__name__, self.item_type)
         else:
             return '%s(name=%r)' % (self.__class__.__name__, self.prefixed_name)
 
@@ -813,31 +832,31 @@ class XsdList(XsdSimpleType):
             if value.tag == XSD_SIMPLE_TYPE:
                 for child in value:
                     if child.tag == XSD_LIST:
-                        super(XsdList, self).__setattr__(name, child)
+                        super().__setattr__(name, child)
                         return
             raise XMLSchemaValueError(
-                "a {0!r} definition required for {1!r}".format(XSD_LIST, self)
+                f"a {XSD_LIST!r} definition required for {self!r}"
             )
-        elif name == 'base_type':
+        elif name == 'item_type':
             if not value.is_atomic():
                 raise XMLSchemaValueError(
-                    _("%r: a list must be based on atomic data types") % self
+                    _("%r: a list must be based on atomic data types") % value
                 )
         elif name == 'white_space' and value is None:
             value = 'collapse'
-        super(XsdList, self).__setattr__(name, value)
+        super().__setattr__(name, value)
 
     def _parse(self) -> None:
-        base_type: Any
+        item_type: Any
 
         child = self._parse_child_component(self.elem)
         if child is not None:
             # Case of a local simpleType declaration inside the list tag
             try:
-                base_type = self.schema.simple_type_factory(child, parent=self)
+                item_type = self.schema.simple_type_factory(child, parent=self)
             except XMLSchemaParseError as err:
                 self.parse_error(err)
-                base_type = self.any_atomic_type
+                item_type = self.any_atomic_type
 
             if 'itemType' in self.elem.attrib:
                 self.parse_error(_("ambiguous list type declaration"))
@@ -851,35 +870,35 @@ class XsdList(XsdSimpleType):
                     self.parse_error(_("missing list type declaration"))
                 else:
                     self.parse_error(err)
-                base_type = self.any_atomic_type
+                item_type = self.any_atomic_type
             else:
                 try:
-                    base_type = self.maps.lookup_type(item_qname)
+                    item_type = self.maps.lookup_type(item_qname)
                 except KeyError:
                     msg = _("unknown type {!r}")
                     self.parse_error(msg.format(self.elem.attrib['itemType']))
-                    base_type = self.any_atomic_type
+                    item_type = self.any_atomic_type
                 else:
-                    if isinstance(base_type, tuple):
+                    if isinstance(item_type, tuple):
                         msg = _("circular definition found for type {!r}")
                         self.parse_error(msg.format(item_qname))
-                        base_type = self.any_atomic_type
+                        item_type = self.any_atomic_type
 
-        if base_type.final == '#all' or 'list' in base_type.final:
+        if item_type.final == '#all' or 'list' in item_type.final:
             msg = _("'final' value of the itemType %r forbids derivation by list")
-            self.parse_error(msg % base_type)
+            self.parse_error(msg % item_type)
 
-        if base_type.name == XSD_ANY_ATOMIC_TYPE:
+        if item_type.name == XSD_ANY_ATOMIC_TYPE:
             msg = _("cannot use xs:anyAtomicType as base type of a user-defined type")
             self.parse_error(msg)
 
         try:
-            self.base_type = base_type
+            self.item_type = item_type
         except XMLSchemaValueError as err:
             self.parse_error(err)
-            self.base_type = self.any_atomic_type
+            self.item_type = self.any_atomic_type
         else:
-            if not base_type.allow_empty and self.min_length != 0:
+            if not item_type.allow_empty and self.min_length != 0:
                 self.allow_empty = False
 
     @property
@@ -891,8 +910,8 @@ class XsdList(XsdSimpleType):
         return XSD_10_LIST_FACETS if self.xsd_version == '1.0' else XSD_11_LIST_FACETS
 
     @property
-    def item_type(self) -> BaseXsdType:
-        return self.base_type
+    def root_type(self) -> BaseXsdType:
+        return self.item_type.root_type
 
     def is_atomic(self) -> bool:
         return False
@@ -911,7 +930,7 @@ class XsdList(XsdSimpleType):
             return False
         elif other.name in self._special_types:
             return derivation != 'extension'
-        elif self.base_type is other:
+        elif self.item_type is other:
             return True
         else:
             return False
@@ -920,8 +939,8 @@ class XsdList(XsdSimpleType):
             -> Iterator[XsdComponent]:
         if xsd_classes is None or isinstance(self, xsd_classes):
             yield self
-        if self.base_type.parent is not None:
-            yield from self.base_type.iter_components(xsd_classes)
+        if self.item_type.parent is not None:
+            yield from self.item_type.iter_components(xsd_classes)
 
     def iter_decode(self, obj: Union[str, bytes],
                     validation: str = 'lax', **kwargs: Any) \
@@ -929,7 +948,7 @@ class XsdList(XsdSimpleType):
                               List[Optional[AtomicValueType]]]]:
         items = []
         for chunk in self.normalize(obj).split():
-            for result in self.base_type.iter_decode(chunk, validation, **kwargs):
+            for result in self.item_type.iter_decode(chunk, validation, **kwargs):
                 if isinstance(result, XMLSchemaValidationError):
                     yield result
                 else:
@@ -945,7 +964,7 @@ class XsdList(XsdSimpleType):
 
         encoded_items: List[Any] = []
         for item in obj:
-            for result in self.base_type.iter_encode(item, validation, **kwargs):
+            for result in self.item_type.iter_encode(item, validation, **kwargs):
                 if isinstance(result, XMLSchemaValidationError):
                     yield result
                 else:
@@ -974,7 +993,7 @@ class XsdUnion(XsdSimpleType):
                  schema: SchemaType,
                  parent: Optional[XsdComponent],
                  name: Optional[str] = None) -> None:
-        super(XsdUnion, self).__init__(elem, schema, parent, name, facets=None)
+        super().__init__(elem, schema, parent, name, facets=None)
 
     def __repr__(self) -> str:
         if self.name is None:
@@ -987,10 +1006,10 @@ class XsdUnion(XsdSimpleType):
             if value.tag == XSD_SIMPLE_TYPE:
                 for child in value:
                     if child.tag == XSD_UNION:
-                        super(XsdUnion, self).__setattr__(name, child)
+                        super().__setattr__(name, child)
                         return
             raise XMLSchemaValueError(
-                "a {0!r} definition required for {1!r}".format(XSD_UNION, self)
+                f"a {XSD_UNION!r} definition required for {self!r}"
             )
 
         elif name == 'white_space':
@@ -998,7 +1017,7 @@ class XsdUnion(XsdSimpleType):
                 msg = _("wrong value %r for attribute 'white_space'")
                 raise XMLSchemaValueError(msg % value)
             value = 'collapse'
-        super(XsdUnion, self).__setattr__(name, value)
+        super().__setattr__(name, value)
 
     def _parse(self) -> None:
         mt: Any
@@ -1109,7 +1128,8 @@ class XsdUnion(XsdSimpleType):
 
         if not isinstance(obj, str) or ' ' not in obj.strip():
             reason = _("invalid value {!r}").format(obj)
-            yield XMLSchemaDecodeError(self, obj, self.member_types, reason)
+            error = XMLSchemaDecodeError(self, obj, self.member_types, reason)
+            yield self.validation_error(validation, error, **kwargs)
             return
 
         items = []
@@ -1131,7 +1151,8 @@ class XsdUnion(XsdSimpleType):
 
         if not_decodable:
             reason = _("no type suitable for decoding the values %r") % not_decodable
-            yield XMLSchemaDecodeError(self, obj, self.member_types, reason)
+            error = XMLSchemaDecodeError(self, obj, self.member_types, reason)
+            yield self.validation_error(validation, error, **kwargs)
 
         yield items if len(items) > 1 else items[0] if items else None
 
@@ -1165,7 +1186,8 @@ class XsdUnion(XsdSimpleType):
 
         if validation != 'skip':
             reason = _("no type suitable for encoding the object")
-            yield XMLSchemaEncodeError(self, obj, self.member_types, reason)
+            error = XMLSchemaEncodeError(self, obj, self.member_types, reason)
+            yield self.validation_error(validation, error, **kwargs)
             yield None
         else:
             yield str(obj)
@@ -1201,7 +1223,7 @@ class XsdAtomicRestriction(XsdAtomic):
                     raise XMLSchemaValueError(
                         "an xs:restriction definition required for %r." % self
                     )
-        super(XsdAtomicRestriction, self).__setattr__(name, value)
+        super().__setattr__(name, value)
 
     def _parse(self) -> None:
         elem = self.elem
@@ -1451,11 +1473,17 @@ class XsdAtomicRestriction(XsdAtomic):
                         except (ValueError, DecimalException, TypeError):
                             pass
 
-                    for validator in self.validators:
-                        try:
-                            validator(obj)
-                        except XMLSchemaValidationError as err:
-                            yield err
+                    if self.validators:
+                        if self.root_type.name in (XSD_QNAME, XSD_NOTATION):
+                            value = get_extended_qname(obj, kwargs.get('namespaces'))
+                        else:
+                            value = obj
+
+                        for validator in self.validators:
+                            try:
+                                validator(value)
+                            except XMLSchemaValidationError as err:
+                                yield err
 
                 if self.patterns:
                     if not isinstance(self.primitive_type, XsdUnion):
