@@ -10,8 +10,8 @@
 """
 This module contains classes for XML Schema wildcards.
 """
-from typing import cast, Any, Callable, Dict, Iterable, Iterator, List, Optional, \
-    Tuple, Union, Counter
+from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, \
+    Tuple, Union
 
 from elementpath import SchemaElementNode, build_schema_node_tree
 
@@ -20,18 +20,15 @@ from ..names import XSI_NAMESPACE, XSD_ANY, XSD_ANY_ATTRIBUTE, \
     XSD_OPEN_CONTENT, XSD_DEFAULT_OPEN_CONTENT, XSI_TYPE
 from ..aliases import ElementType, SchemaType, SchemaElementType, SchemaAttributeType, \
     ModelGroupType, ModelParticleType, AtomicValueType, IterDecodeType, IterEncodeType, \
-    DecodedValueType, EncodedValueType
+    DecodedValueType, EncodedValueType, OccursCounterType
 from ..translation import gettext as _
 from ..helpers import get_namespace, raw_xml_encode
-from ..xpath import XsdSchemaProtocol, XsdElementProtocol, XMLSchemaProxy, ElementPathMixin
+from ..xpath import XMLSchemaProxy, ElementPathMixin
 from .xsdbase import ValidationMixin, XsdComponent
 from .particles import ParticleMixin
 from . import elements
 
 
-OccursCounterType = Counter[Union[ModelParticleType, Tuple[ModelParticleType]]]
-
-
 class XsdWildcard(XsdComponent):
     names = ()
     namespace: Union[Tuple[str], List[str]] = ('##any',)
@@ -166,8 +163,7 @@ class XsdWildcard(XsdComponent):
         elif not default_namespace:
             return self.is_namespace_allowed('')
         else:
-            return self.is_namespace_allowed('') or \
-                self.is_namespace_allowed(default_namespace)
+            return self.is_namespace_allowed(default_namespace)
 
     def is_namespace_allowed(self, namespace: str) -> bool:
         if self.not_namespace:
@@ -206,12 +202,14 @@ class XsdWildcard(XsdComponent):
 
     def is_restriction(self, other: Union[ModelParticleType, 'XsdAnyAttribute'],
                        check_occurs: bool = True) -> bool:
-        if check_occurs and isinstance(self, ParticleMixin) \
-                and not isinstance(other, XsdAnyAttribute) \
-                and not self.has_occurs_restriction(other):
-            return False
-        elif not isinstance(other, self.__class__):
+        if not isinstance(other, self.__class__):
             return False
+        elif check_occurs and isinstance(self, ParticleMixin):
+            if not isinstance(other, XsdAnyAttribute) and \
+                    not self.has_occurs_restriction(other):
+                return False
+            elif self.max_occurs == 0:
+                return True
 
         other: XsdWildcard  # type: ignore[no-redef]
         if other.process_contents == 'strict' and self.process_contents != 'strict':
@@ -408,7 +406,7 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
 
     def __init__(self, elem: ElementType, schema: SchemaType, parent: XsdComponent) -> None:
         self.precedences = {}
-        super(XsdAnyElement, self).__init__(elem, schema, parent)
+        super().__init__(elem, schema, parent)
 
     def __repr__(self) -> str:
         if self.namespace:
@@ -424,26 +422,23 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
 
     @property
     def xpath_proxy(self) -> XMLSchemaProxy:
-        return XMLSchemaProxy(
-            schema=cast(XsdSchemaProtocol, self.schema),
-            base_element=cast(XsdElementProtocol, self)
-        )
+        return XMLSchemaProxy(self.schema, self)
 
     @property
     def xpath_node(self) -> SchemaElementNode:
         schema_node = self.schema.xpath_node
-        node = schema_node.get_element_node(cast(XsdElementProtocol, self))
+        node = schema_node.get_element_node(self)
         if isinstance(node, SchemaElementNode):
             return node
 
         return build_schema_node_tree(
-            root=cast(XsdElementProtocol, self),
+            root=self,
             elements=schema_node.elements,
             global_elements=schema_node.children,
         )
 
     def _parse(self) -> None:
-        super(XsdAnyElement, self)._parse()
+        super()._parse()
         self._parse_particle(self.elem)
 
     def match(self, name: Optional[str], default_namespace: Optional[str] = None,
@@ -466,7 +461,7 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
 
         try:
             if name[0] != '{' and default_namespace:
-                return self.maps.lookup_element('{%s}%s' % (default_namespace, name))
+                return self.maps.lookup_element(f'{{{default_namespace}}}{name}')
             else:
                 return self.maps.lookup_element(name)
         except LookupError:
@@ -497,7 +492,7 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
                 return
 
         namespace = get_namespace(obj.tag)
-        if not self.maps.load_namespace(namespace):
+        if namespace not in self.maps.namespaces and not self.maps.load_namespace(namespace):
             reason = f"unavailable namespace {namespace!r}"
         else:
             try:
@@ -523,7 +518,11 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
 
         if validation != 'skip' and self.process_contents == 'strict':
             yield self.validation_error(validation, reason, obj, **kwargs)
-        yield from self.any_type.iter_decode(obj, validation, **kwargs)
+
+        xsd_element = self.maps.validator.create_element(
+            obj.tag, parent=self, form='unqualified'
+        )
+        yield from xsd_element.iter_decode(obj, validation, **kwargs)
 
     def iter_encode(self, obj: Tuple[str, ElementType], validation: str = 'lax', **kwargs: Any) \
             -> IterEncodeType[Any]:
@@ -549,7 +548,7 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
                 yield from xsd_element.iter_encode(value, validation, **kwargs)
                 return
 
-        # Check if there is an xsi:type attribute, but it has to extract
+        # Check if there is a xsi:type attribute, but it has to extract
         # attributes using the converter instance.
         if self.process_contents == 'strict':
             xsd_element = self.maps.validator.create_element(
@@ -563,12 +562,12 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
         try:
             converter = kwargs['converter']
         except KeyError:
-            converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+            converter = self._get_converter(value, kwargs)
 
         try:
             level = kwargs['level']
         except KeyError:
-            level = 0
+            level = kwargs['level'] = 0
 
         try:
             element_data = converter.element_encode(value, xsd_element, level)
@@ -582,6 +581,7 @@ class XsdAnyElement(XsdWildcard, ParticleMixin,
 
         if validation != 'skip' and self.process_contents == 'strict':
             yield self.validation_error(validation, reason, **kwargs)
+
         yield from self.any_type.iter_encode(obj, validation, **kwargs)
 
     def is_overlap(self, other: ModelParticleType) -> bool:
@@ -660,7 +660,7 @@ class XsdAnyAttribute(XsdWildcard, ValidationMixin[Tuple[str, str], DecodedValue
 
         try:
             if name[0] != '{' and default_namespace:
-                return self.maps.lookup_attribute('{%s}%s' % (default_namespace, name))
+                return self.maps.lookup_attribute(f'{{{default_namespace}}}{name}')
             else:
                 return self.maps.lookup_attribute(name)
         except LookupError:
@@ -743,7 +743,7 @@ class Xsd11AnyElement(XsdAnyElement):
         </any>
     """
     def _parse(self) -> None:
-        super(Xsd11AnyElement, self)._parse()
+        super()._parse()
         self._parse_not_constraints()
 
     def is_matching(self, name: Optional[str],
@@ -753,10 +753,10 @@ class Xsd11AnyElement(XsdAnyElement):
                     **kwargs: Any) -> bool:
         """
         Returns `True` if the component name is matching the name provided as argument,
-        `False` otherwise. For XSD elements the matching is extended to substitutes.
+        `False` otherwise.
 
         :param name: a local or fully-qualified name.
-        :param default_namespace: used if it's not None and not empty for completing \
+        :param default_namespace: used by the XPath processor for completing \
         the name argument in case it's a local name.
         :param group: used only by XSD 1.1 any element wildcards to verify siblings in \
         case of ##definedSibling value in notQName attribute.
@@ -771,16 +771,15 @@ class Xsd11AnyElement(XsdAnyElement):
             if not self.is_namespace_allowed(''):
                 return False
         else:
-            name = '{%s}%s' % (default_namespace, name)
-            if not self.is_namespace_allowed('') \
-                    and not self.is_namespace_allowed(default_namespace):
+            name = f'{{{default_namespace}}}{name}'
+            if not self.is_namespace_allowed(default_namespace):
                 return False
 
         if group in self.precedences:
             if occurs is None:
                 if any(e.is_matching(name) for e in self.precedences[group]):
                     return False
-            elif any(e.is_matching(name) and not e.is_over(occurs[e])
+            elif any(e.is_matching(name) and not e.is_over(occurs)
                      for e in self.precedences[group]):
                 return False
 
@@ -821,7 +820,7 @@ class Xsd11AnyAttribute(XsdAnyAttribute):
         </anyAttribute>
     """
     def _parse(self) -> None:
-        super(Xsd11AnyAttribute, self)._parse()
+        super()._parse()
         self._parse_not_constraints()
 
     def is_matching(self, name: Optional[str],
@@ -834,7 +833,7 @@ class Xsd11AnyAttribute(XsdAnyAttribute):
         elif not default_namespace:
             namespace = ''
         else:
-            name = '{%s}%s' % (default_namespace, name)
+            name = f'{{{default_namespace}}}{name}'
             namespace = default_namespace
 
         if '##defined' in self.not_qname and name in self.maps.attributes:
@@ -864,19 +863,19 @@ class XsdOpenContent(XsdComponent):
     any_element = None  # type: Xsd11AnyElement
 
     def __init__(self, elem: ElementType, schema: SchemaType, parent: XsdComponent) -> None:
-        super(XsdOpenContent, self).__init__(elem, schema, parent)
+        super().__init__(elem, schema, parent)
 
     def __repr__(self) -> str:
         return '%s(mode=%r)' % (self.__class__.__name__, self.mode)
 
     def _parse(self) -> None:
-        super(XsdOpenContent, self)._parse()
+        super()._parse()
         try:
             self.mode = self.elem.attrib['mode']
         except KeyError:
             pass
         else:
-            if self.mode not in {'none', 'interleave', 'suffix'}:
+            if self.mode not in ('none', 'interleave', 'suffix'):
                 msg = _("wrong value %r for 'mode' attribute")
                 self.parse_error(msg % self.mode)
 
@@ -889,10 +888,7 @@ class XsdOpenContent(XsdComponent):
         elif child is None or child.tag != XSD_ANY:
             self.parse_error(_("an <xs:any> child declaration is required"))
         else:
-            any_element = Xsd11AnyElement(child, self.schema, self)
-            any_element.min_occurs = 0
-            any_element.max_occurs = None
-            self.any_element = any_element
+            self.any_element = Xsd11AnyElement(child, self.schema, self)
 
     @property
     def built(self) -> bool:
@@ -926,7 +922,7 @@ class XsdDefaultOpenContent(XsdOpenContent):
         super(XsdOpenContent, self).__init__(elem, schema)
 
     def _parse(self) -> None:
-        super(XsdDefaultOpenContent, self)._parse()
+        super()._parse()
         if self.parent is not None:
             msg = _("defaultOpenContent must be a child of the schema")
             self.parse_error(msg)
@@ -938,5 +934,5 @@ class XsdDefaultOpenContent(XsdOpenContent):
             self.parse_error(msg)
 
         if 'appliesToEmpty' in self.elem.attrib:
-            if self.elem.attrib['appliesToEmpty'].strip() in {'true', '1'}:
+            if self.elem.attrib['appliesToEmpty'].strip() in ('true', '1'):
                 self.applies_to_empty = True
